from __future__ import annotations


from typing import Self

import moderngl
import numpy as np

from ...constants.custom_typing import ShapeType
from ...lazy.lazy import Lazy
from ...toplevel.toplevel import Toplevel
from ..field import (
    Field,
    StructuredField
)
from .buffer import Buffer


class UniformBlockBuffer(Buffer):
    __slots__ = ()

    def __init__(
        self: Self,
        *,
        name: str,
        field_declarations: tuple[str, ...],
        structs: dict[str, tuple[str, ...]] | None = None,
        data_dict: dict[str, np.ndarray],
        array_lens: dict[str, int] | None = None
    ) -> None:
        super().__init__(
            array_lens=array_lens
        )
        self._name_ = name
        self._field_declarations_ = field_declarations
        if structs is not None:
            self._struct_items_ = tuple(structs.items())
        self._data_dict_ = data_dict

    @Lazy.variable()
    @staticmethod
    def _name_() -> str:
        return ""

    @Lazy.variable(plural=True)
    @staticmethod
    def _field_declarations_() -> tuple[str, ...]:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _struct_items_() -> tuple[tuple[str, tuple[str, ...]], ...]:
        return ()

    @Lazy.variable()
    @staticmethod
    def _data_dict_() -> dict[str, np.ndarray]:
        return {}

    @Lazy.property()
    @staticmethod
    def _field_(
        name: str,
        field_declarations: tuple[str, ...],
        struct_items: tuple[tuple[str, tuple[str, ...]], ...],
        array_len_items: tuple[tuple[str, int], ...]
    ) -> StructuredField:
        return StructuredField(
            name=name,
            shape=(),
            fields=tuple(
                Field.parse_field(
                    field_declaration=field_declaration,
                    struct_dict=dict(struct_items),
                    array_lens_dict=dict(array_len_items)
                )
                for field_declaration in field_declarations
            )
        )

    @Lazy.property()
    @staticmethod
    def _buffer_(
        field: StructuredField,
        shape: ShapeType,
        data_dict: dict[str, np.ndarray]
    ) -> moderngl.Buffer:
        return Toplevel._get_context().buffer(data=field.write(shape, data_dict))



from __future__ import annotations


from typing import (
    Iterator,
    Self
)

import moderngl
import pyglet.gl as gl

from ..rendering.mgl_enums import (
    BlendEquation,
    BlendFunc,
    ContextFlag,
    PrimitiveMode
)
from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


class Context(ToplevelResource):
    __slots__ = ("_mgl_context",)

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        mgl_context = moderngl.create_context(
            require=Toplevel._get_config().gl_version_code
        )
        mgl_context.gc_mode = "auto"
        self._mgl_context: moderngl.Context = mgl_context

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        Toplevel._context = self
        yield
        Toplevel._context = None

    def set_blendings(
        self: Self,
        blendings: tuple[tuple[BlendFunc, BlendFunc, BlendEquation], ...]
    ) -> None:
        for index, (src_blend_func, dst_blend_func, blend_equation) in enumerate(blendings):
            gl.glBlendFunci(
                index,
                src_blend_func.value,
                dst_blend_func.value
            )
            gl.glBlendEquationi(
                index,
                blend_equation.value
            )

    def set_flag(
        self: Self,
        flag: ContextFlag
    ) -> None:
        self._mgl_context.enable_only(flag.value)

    @property
    def version_code(
        self: Self
    ) -> int:
        return self._mgl_context.version_code

    @property
    def screen_framebuffer(
        self: Self
    ) -> moderngl.Framebuffer:
        return self._mgl_context.screen

    def texture(
        self: Self,
        *,
        size: tuple[int, int],
        components: int,
        samples: int,
        dtype: str
    ) -> moderngl.Texture:
        return self._mgl_context.texture(
            size=size,
            components=components,
            samples=samples,
            dtype=dtype
        )

    def framebuffer(
        self: Self,
        *,
        color_attachments: tuple[moderngl.Texture, ...]
    ) -> moderngl.Framebuffer:
        return self._mgl_context.framebuffer(
            color_attachments=color_attachments
        )

    def buffer(
        self: Self,
        *,
        data: bytes
    ) -> moderngl.Buffer:
        return self._mgl_context.buffer(data=data)

    def program(
        self: Self,
        *,
        vertex_shader: str,
        fragment_shader: str | None = None,
        geometry_shader: str | None = None,
        tess_control_shader: str | None = None,
        tess_evaluation_shader: str | None = None,
        varyings: tuple[str, ...] = ()
    ) -> moderngl.Program:
        return self._mgl_context.program(
            vertex_shader=vertex_shader,
            fragment_shader=fragment_shader,
            geometry_shader=geometry_shader,
            tess_control_shader=tess_control_shader,
            tess_evaluation_shader=tess_evaluation_shader,
            varyings=varyings
        )

    def vertex_array(
        self: Self,
        *,
        program: moderngl.Program,
        attributes_buffer: moderngl.Buffer,
        attributes_buffer_format_str: str,
        attribute_names: tuple[str, ...],
        index_buffer: moderngl.Buffer | None,
        mode: PrimitiveMode
    ) -> moderngl.VertexArray:
        content = []
        if attribute_names:
            content.append((attributes_buffer, attributes_buffer_format_str, *attribute_names))
        return self._mgl_context.vertex_array(
            program=program,
            content=content,
            index_buffer=index_buffer,
            index_element_size=4,
            mode=mode.value
        )

    def scope(
        self: Self,
        *,
        framebuffer: moderngl.Framebuffer | None = None,
        textures: tuple[tuple[moderngl.Texture, int], ...] = (),
        uniform_buffers: tuple[tuple[moderngl.Buffer, int], ...] = ()
    ) -> moderngl.Scope:
        return self._mgl_context.scope(
            framebuffer=framebuffer,
            textures=textures,
            uniform_buffers=uniform_buffers
        )

    def copy_framebuffer(
        self: Self,
        *,
        dst: moderngl.Framebuffer,
        src: moderngl.Framebuffer
    ) -> None:
        self._mgl_context.copy_framebuffer(
            dst=dst,
            src=src
        )

    def blit_framebuffer(
        self: Self,
        *,
        dst: moderngl.Framebuffer,
        src: moderngl.Framebuffer
    ) -> None:
        # Moderngl `copy_framebuffer` does not support unequal frame sizes.
        # This method is a simple patch for such requirement.
        gl.glBindFramebuffer(gl.GL_READ_FRAMEBUFFER, src.glo)
        gl.glBindFramebuffer(gl.GL_DRAW_FRAMEBUFFER, dst.glo)
        gl.glBlitFramebuffer(
            *src.viewport, *dst.viewport,
            gl.GL_COLOR_BUFFER_BIT, gl.GL_LINEAR
        )



from __future__ import annotations


import pathlib
import re
from typing import (
    ClassVar,
    Iterator,
    Literal,
    Self,
    TypedDict
)

import attrs
import cairocffi
import pangocairocffi
import pangocffi

from ...animatables.arrays.animatable_color import AnimatableColor
from ...constants.custom_typing import (
    AlignmentType,
    ColorType
)
from ...toplevel.toplevel import Toplevel
from .string_mobject import (
    CommandInfo,
    StandaloneCommandInfo,
    StringMobjectIO,
    StringMobjectInput,
    StringMobjectKwargs
)


# See `https://docs.gtk.org/Pango/pango_markup.html`.
class PangoAttributes(TypedDict, total=False):
    font: str
    font_desc: str
    font_family: str
    face: str
    font_size: str
    size: str
    font_style: Literal["normal", "oblique", "italic"]
    style: Literal["normal", "oblique", "italic"]
    font_weight: Literal["ultralight", "light", "normal", "bold", "ultrabold", "heavy"] | str
    weight: Literal["ultralight", "light", "normal", "bold", "ultrabold", "heavy"] | str
    font_variant: Literal["normal", "small-caps", "all-small-caps", "petite-caps", "all-petite-caps", "unicase", "title-caps"]
    variant: Literal["normal", "small-caps", "all-small-caps", "petite-caps", "all-petite-caps", "unicase", "title-caps"]
    font_stretch: str
    stretch: str
    font_features: str
    foreground: str
    fgcolor: str
    color: str
    background: str
    bgcolor: str
    alpha: str
    fgalpha: str
    background_alpha: str
    bgalpha: str
    underline: Literal["none", "single", "double", "low", "error"]
    underline_color: str
    overline: Literal["none", "single"]
    overline_color: str
    rise: str
    baseline_shift: str
    font_scale: Literal["superscript", "subscript", "small-caps"]
    strikethrough: Literal["true", "false"]
    strikethrough_color: str
    fallback: Literal["true", "false"]
    lang: str
    letter_spacing: str
    gravity: Literal["south", "east", "north", "west", "auto"]
    gravity_hint: Literal["natural", "strong", "line"]
    show: Literal["none", "spaces", "line-breaks", "spaces|line-breaks", "ignorables", "spaces|ignorables", "line-breaks|ignorables", "spaces|line-breaks|ignorables"]
    insert_hyphens: Literal["true", "false"]
    allow_breaks: Literal["true", "false"]
    line_height: str
    text_transform: Literal["none", "lowercase", "uppercase", "capitalize"]
    segment: Literal["word", "sentence"]


@attrs.frozen(kw_only=True)
class PangoStringMobjectInput(StringMobjectInput[PangoAttributes]):
    color: ColorType = attrs.field(factory=lambda: Toplevel._get_config().default_color)
    alignment: AlignmentType = attrs.field(factory=lambda: Toplevel._get_config().pango_alignment)
    font: str = attrs.field(factory=lambda: Toplevel._get_config().pango_font)


class PangoStringMobjectKwargs(StringMobjectKwargs[PangoAttributes], total=False):
    color: ColorType
    alignment: AlignmentType
    font: str


class PangoStringMobjectIO[PangoStringMobjectInputT: PangoStringMobjectInput](StringMobjectIO[PangoAttributes, PangoStringMobjectInputT]):
    __slots__ = ()

    _MARKUP_TAGS: ClassVar[dict[str, PangoAttributes]] = {
        "b": PangoAttributes(font_weight="bold"),
        "big": PangoAttributes(font_size="larger"),
        "i": PangoAttributes(font_style="italic"),
        "s": PangoAttributes(strikethrough="true"),
        "sub": PangoAttributes(baseline_shift="subscript", font_scale="subscript"),
        "sup": PangoAttributes(baseline_shift="superscript", font_scale="superscript"),
        "small": PangoAttributes(font_size="smaller"),
        "tt": PangoAttributes(font_family="monospace"),
        "u": PangoAttributes(underline="single")
    }
    _MARKUP_ESCAPE_DICT: ClassVar[dict[str, str]] = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        "\"": "&quot;",
        "'": "&apos;"
    }

    @classmethod
    def _create_svg(
        cls: type[Self],
        content: str,
        input_data: PangoStringMobjectInputT,
        svg_path: pathlib.Path
    ) -> None:
        match input_data.alignment:
            case "left":
                pango_alignment = pangocffi.Alignment.LEFT
            case "center":
                pango_alignment = pangocffi.Alignment.CENTER
            case "right":
                pango_alignment = pangocffi.Alignment.RIGHT

        with svg_path.open(mode="wb") as svg_file:
            surface = cairocffi.SVGSurface(svg_file, 1024.0, 1024.0)
            context = cairocffi.Context(surface)
            layout = pangocairocffi.create_layout(context)
            layout.alignment = pango_alignment
            layout.apply_markup(content)
            pangocairocffi.show_layout(context, layout)
            surface.finish()

    @classmethod
    def _get_adjustment_scale(
        cls: type[Self]
    ) -> float:
        return 0.05626

    @classmethod
    def _iter_global_span_attributes(
        cls: type[Self],
        input_data: PangoStringMobjectInputT,
        temp_path: pathlib.Path
    ) -> Iterator[PangoAttributes]:
        yield PangoAttributes(
            foreground=AnimatableColor._color_to_hex(input_data.color),
            font_family=input_data.font
        )
        yield from super()._iter_global_span_attributes(input_data, temp_path)

    @classmethod
    def _get_empty_attributes(
        cls: type[Self]
    ) -> PangoAttributes:
        return PangoAttributes()

    @classmethod
    def _get_command_pair(
        cls: type[Self],
        attributes: PangoAttributes
    ) -> tuple[str, str]:
        return f"<span {" ".join(
            f"{key}='{value}'"
            for key, value in attributes.items()
        )}>", "</span>"

    @classmethod
    def _convert_attributes_for_labelling(
        cls: type[Self],
        attributes: PangoAttributes,
        label: int | None
    ) -> PangoAttributes:
        result = PangoAttributes(attributes)
        for key in (
            "foreground",
            "fgcolor",
            "color"
        ):
            if key in result:
                result.pop(key)

        for key in (
            "background",
            "bgcolor",
            "underline_color",
            "overline_color",
            "strikethrough_color"
        ):
            if key in result:
                result[key] = "black"

        if label is not None:
            result["foreground"] = f"#{label:06x}"
        return result

    @classmethod
    def _iter_command_infos(
        cls: type[Self],
        string: str
    ) -> Iterator[CommandInfo[PangoAttributes]]:
        pattern = re.compile(r"""[<>&"']""")
        for match in pattern.finditer(string):
            yield StandaloneCommandInfo(match, replacement=cls._markup_escape(match.group()))

    @classmethod
    def _markup_escape(
        cls: type[Self],
        string: str
    ) -> str:
        return cls._MARKUP_ESCAPE_DICT.get(string, string)



from __future__ import annotations


from typing import (
    Iterator,
    Self
)

from ...animatables.animatable.animatable import AnimatableActions
from ...animatables.arrays.animatable_color import AnimatableColor
from ...animatables.arrays.animatable_float import AnimatableFloat
from ...animatables.graph import Graph
from ...animatables.model import ModelActions
from ...constants.custom_typing import (
    NP_3f8,
    NP_f8,
    NP_x2i4,
    NP_x3f8
)
from ...lazy.lazy import Lazy
from ...rendering.buffers.attributes_buffer import AttributesBuffer
from ...rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from ...rendering.mgl_enums import PrimitiveMode
from ...rendering.vertex_array import VertexArray
from ...toplevel.toplevel import Toplevel
from ..mobject import Mobject


class GraphMobject(Mobject):
    __slots__ = ()

    def __init__(
        self: Self,
        graph: Graph | None = None
    ) -> None:
        super().__init__()
        if graph is not None:
            self._graph_ = graph

    @AnimatableActions.interpolate.register_descriptor()
    @AnimatableActions.piecewise.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _graph_() -> Graph:
        return Graph()

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableColor)
    @Lazy.volatile()
    @staticmethod
    def _color_() -> AnimatableColor:
        return AnimatableColor(Toplevel._get_config().default_color)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _opacity_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().default_opacity)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _weight_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().default_weight)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _width_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().graph_width)

    @Lazy.property()
    @staticmethod
    def _local_sample_positions_(
        graph__positions: NP_x3f8,
        graph__edges: NP_x2i4
    ) -> NP_x3f8:
        return graph__positions[graph__edges.flatten()]

    @Lazy.property()
    @staticmethod
    def _graph_uniform_block_buffer_(
        color__array: NP_3f8,
        opacity__array: NP_f8,
        weight__array: NP_f8,
        width__array: NP_f8
    ) -> UniformBlockBuffer:
        return UniformBlockBuffer(
            name="ub_graph",
            field_declarations=(
                "vec3 u_color",
                "float u_opacity",
                "float u_weight",
                "float u_width"
            ),
            data_dict={
                "u_color": color__array,
                "u_opacity": opacity__array,
                "u_weight": weight__array,
                "u_width": width__array
            }
        )

    @Lazy.property()
    @staticmethod
    def _graph_attributes_buffer_(
        graph__positions: NP_x3f8,
        graph__edges: NP_x2i4
    ) -> AttributesBuffer:
        return AttributesBuffer(
            field_declarations=(
                "vec3 in_position",
            ),
            data_dict={
                "in_position": graph__positions
            },
            index=graph__edges.flatten(),
            primitive_mode=PrimitiveMode.LINES,
            vertices_count=len(graph__positions)
        )

    @Lazy.property()
    @staticmethod
    def _graph_vertex_array_(
        camera__camera_uniform_block_buffer: UniformBlockBuffer,
        model_uniform_block_buffer: UniformBlockBuffer,
        graph_uniform_block_buffer: UniformBlockBuffer,
        graph_attributes_buffer: AttributesBuffer
    ) -> VertexArray:
        return VertexArray(
            shader_filename="graph.glsl",
            uniform_block_buffers=(
                camera__camera_uniform_block_buffer,
                model_uniform_block_buffer,
                graph_uniform_block_buffer
            ),
            attributes_buffer=graph_attributes_buffer
        )

    def _iter_vertex_arrays(
        self: Self
    ) -> Iterator[VertexArray]:
        yield from super()._iter_vertex_arrays()
        yield self._graph_vertex_array_



import re
from typing import Self

import numpy as np
from manim3 import *


class ShapeTransformExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        square = (
            Square()
            .set(color=WHITE, opacity=1.0)
            .add_strokes(color=YELLOW, width=0.0)
        )
        circle = (
            Circle()
            .set(color=PINK, opacity=0.9)
            .add_strokes(color=YELLOW, weight=10.0)
        )
        triangle = (
            RegularPolygon(3)
            .rotate_about_origin(OUT * (PI / 2.0))
            .set(color=BLUE, opacity=0.9)
            .add_strokes(color=GREEN_A, weight=10.0)
        )

        self.add(square)
        await self.wait()
        await self.play(Transform(square, circle), run_time=2.0, rate=Rates.smooth())
        await self.wait()
        await self.play(Transform(circle, triangle), run_time=2.0, rate=Rates.smooth())
        await self.wait()


class TextTransformExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        text = (
            Text("Text", concatenate=True)
            .set(color=ORANGE, opacity=0.5)
            .shift(2.0 * LEFT)
            .add_strokes(color=BLUE, weight=10.0)
        )
        tex = (
            Tex("Tex", concatenate=True)
            .set(color=BLUE, opacity=0.5)
            .add_strokes(color=PINK, weight=10.0)
        )
        code = Code("print(\"Code!\")").shift(2.0 * RIGHT)
        self.add(text)
        await self.wait()
        await self.play(Transform(text, tex), run_time=2.0, rate=Rates.smooth())
        await self.wait()
        await self.play(FadeTransform(tex, code), run_time=2.0, rate=Rates.smooth())
        await self.wait(3.0)


class CreateExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        text = (
            Text("Text")
            .set(color=ORANGE, opacity=0.5)
            .add_strokes(color=BLUE, weight=10.0)
        )
        await self.wait()
        await self.play(Create(text), rate=Rates.smooth(), run_time=2.0)
        await self.wait()
        await self.play(Uncreate(text, backwards=True), rate=Rates.smooth(), run_time=2.0)
        await self.wait()


class WriteExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        tex = Tex("Hello").scale(2.0)
        await self.play(Parallel(*(
            Series(
                Create(stroke := glyph.build_stroke()),
                Parallel(
                    FadeIn(glyph),
                    FadeOut(stroke)
                )
            )
            for glyph in tex
            if isinstance(glyph, ShapeMobject)
        ), lag_ratio=0.3), run_time=3.0)
        await self.wait(2.0)


class ThreeDExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        dodec = (
            Dodecahedron()
            .scale(2.0)
            .set(
                color="#00FFAA",
                opacity=0.25
            )
            .bind_lighting(Lighting(
                AmbientLight().set(color=GREY_D),
                PointLight().shift(5.0 * RIGHT)
            ))
        )
        self.add(dodec)
        self.prepare(self.camera.animate(infinite=True).rotate(0.5 * DOWN))

        text = Text("Dodecahedron").scale(0.5)
        await self.play(Parallel(*(
            Parallel(
                FadeIn(char),
                char.animate(rewind=True).shift(DOWN)
            )
            for char in text
        ), lag_time=0.5), rate=Rates.smooth())
        await self.wait(3.0)


class OITExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        self.add(*(
            (Circle()
                .set(color=color, opacity=opacity, weight=weight)
                .shift(0.5 * RIGHT)
                .rotate_about_origin(angle * OUT)
            )
            for color, opacity, weight, angle in zip(
                (RED, GREEN, BLUE),
                (0.3, 0.5, 0.6),
                (1.0, 2.0, 1.0),
                np.linspace(0.0, TAU, 3, endpoint=False),
                strict=True
            )
        ))
        await self.wait(5.0)


class LaggedAnimationExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        text = Text("Text")
        await self.play(Parallel(*(
            Parallel(
                FadeIn(char),
                char.animate(rewind=True).shift(DOWN)
            )
            for char in text
        ), lag_time=0.5), rate=Rates.smooth())
        await self.wait(3.0)


class FormulaExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        factored_formula = MathTex(
            "\\left( a_{0}^{2} + a_{1}^{2} \\right) \\left( b_{0}^{2} + b_{1}^{2} + b_{2}^{2} \\right)",
            local_colors={
                re.compile(r"a_{\d}"): TEAL,
                re.compile(r"b_{\d}"): ORANGE
            }
        ).scale(0.5).shift(UP)
        expanded_formula = MathTex(
            "a_{0}^{2} b_{0}^{2} + a_{0}^{2} b_{1}^{2} + a_{0}^{2} b_{2}^{2}" \
                + " + a_{1}^{2} b_{0}^{2} + a_{1}^{2} b_{1}^{2} + a_{1}^{2} b_{2}^{2}",
            local_colors={
                re.compile(r"a_{\d}"): TEAL,
                re.compile(r"b_{\d}"): ORANGE
            }
        ).scale(0.5).shift(DOWN)
        self.add(factored_formula)
        await self.wait()
        await self.play(TransformMatchingStrings(factored_formula, expanded_formula), rate=Rates.smooth(), run_time=2.0)
        await self.wait(2.0)


class InteractiveExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        squares = ShapeMobject().add(*(
            (
                Circle()
                .scale(0.5)
                .shift(x * RIGHT)
                .set(color=color)
            )
            for x, color in zip(
                np.linspace(-4.0, 4.0, 5),
                (RED, YELLOW, GREEN, BLUE, PURPLE),
                strict=True
            )
        ))
        text = Text("Press space to animate.").shift(1.5 * DOWN)
        self.add(squares, text)
        timelines = [
            square.animate().shift(UP)
            for square in squares
        ]
        for timeline in timelines:
            self.prepare(
                timeline,
                rate=Rates.smooth(),
                launch_condition=KeyPress(KEY.SPACE).captured
            )
        await self.wait_until(lambda: all(timeline.terminated() for timeline in timelines))
        await self.wait()


class NoteTimeline(Timeline):
    __slots__ = (
        "_note",
        "_key"
    )

    def __init__(
        self,
        note: Mobject,
        key: int
    ) -> None:
        super().__init__()
        self._note: Mobject = note
        self._key: int = key

    async def construct(
        self: Self
    ) -> None:
        note = self._note
        self.scene.add(note)
        event = KeyPress(self._key)
        await self.play(
            note.animate(infinite=True).shift(7.0 * DOWN),
            terminate_condition=lambda: event.captured() and -3.4 <= note.box.get()[1] <= -2.6 or note.box.get()[1] <= -3.4
        )
        if event.captured():
            await self.play(
                note.animate().set(opacity=0.0).scale(1.5),
                rate=Rates.rush_from()
            )
        else:
            note.set(opacity=0.4)
            await self.play(
                note.animate(infinite=True).shift(10.0 * DOWN),
                terminate_condition=lambda: note.box.get()[1] <= -5.0
            )
        self.scene.discard(note)


class GameExample(Scene):
    async def construct(
        self: Self
    ) -> None:
        score = [
            "|  | |    |      |     |   | |  ",
            " |  |    |     |   |  |   |  | |",
            "| |   |   | |  || |  |  |   | | ",
            "    |  | |    |  |  |   || |   |"
        ]
        keys = [KEY.D, KEY.F, KEY.J, KEY.K]
        x_coords = np.linspace(-3.0, 3.0, 4)
        note_template = (
            Polygon(np.array((
                (0.5, 0.0),
                (0.46, 0.04),
                (-0.46, 0.04),
                (-0.5, 0.0),
                (-0.46, -0.04),
                (0.46, -0.04)
            )))
            .set(
                width=0.25,
                color=BLUE_B,
                opacity=0.95
            )
            .add_strokes()
        )
        judge_line = Line(8.0 * LEFT, 8.0 * RIGHT).shift(3.0 * DOWN).set(
            width=0.03,
            color=GOLD_A
        )
        key_texts = [
            Text(char, concatenate=True).add_strokes().shift(x_coord * RIGHT + 2.0 * DOWN)
            for char, x_coord in zip("DFJK", x_coords, strict=True)
        ]

        self.add(judge_line)
        await self.play(Parallel(*(
            Create(key_text, n_segments=2)
            for key_text in key_texts
        ), lag_time=0.5), run_time=1.5)
        await self.wait()
        await self.play(Parallel(*(
            Uncreate(key_text)
            for key_text in key_texts
        ), lag_time=0.5), run_time=1.5)
        for note_chars in zip(*score, strict=True):
            for note_char, key, x_coord in zip(note_chars, keys, x_coords, strict=True):
                if note_char == " ":
                    continue
                note = note_template.copy().shift(x_coord * RIGHT + 5.0 * UP)
                self.prepare(NoteTimeline(note, key))
            await self.wait(0.25)
        await self.wait(3.0)


def main() -> None:
    with (
        Config(
            #fps=30,
            #pixel_height=540
            #msaa_samples=0
        ),
        Toplevel.livestream(),
        #Toplevel.recording("ShapeTransformExample.mp4")
    ):
        ShapeTransformExample().run()


if __name__ == "__main__":
    main()



from __future__ import annotations


from abc import (
    ABC,
    abstractmethod
)
from contextlib import (
    AbstractContextManager,
    contextmanager
)
from types import TracebackType
from typing import (
    ClassVar,
    Iterator,
    Self
)


class ToplevelResource(ABC):
    __slots__ = ()

    __context_manager: ClassVar[AbstractContextManager[None] | None] = None

    @abstractmethod
    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        pass

    def __enter__(
        self: Self
    ) -> None:
        cls = type(self)
        assert cls.__context_manager is None
        context_manager = contextmanager(self.__contextmanager__)()
        cls.__context_manager = context_manager
        return context_manager.__enter__()

    def __exit__(
        self: Self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None
    ) -> bool | None:
        cls = type(self)
        context_manager = cls.__context_manager
        assert context_manager is not None
        cls.__context_manager = None
        return context_manager.__exit__(exc_type, exc_value, exc_traceback)



from __future__ import annotations


import itertools
from typing import (
    Iterator,
    Literal,
    Self,
    Unpack
)

import attrs
import mapbox_earcut
import numpy as np
import pyclipr

from ..constants.custom_typing import (
    NP_x2f8,
    NP_x2i4,
    NP_x3f8,
    NP_x3i4,
    NP_xf8,
    NP_xi4
)
from ..lazy.lazy import Lazy
from .animatable.actions import Action
from .animatable.animatable import (
    Animatable,
    AnimatableActions,
    AnimatableInterpolateAnimation,
    AnimatablePiecewiseAnimation,
    DynamicAnimatable
)
from .animatable.animation import (
    AnimateKwargs,
    Animation
)
from .animatable.piecewiser import Piecewiser
from .graph import (
    Graph,
    GraphUtils
)


@attrs.frozen(kw_only=True)
class Triangulation:
    coordinates: NP_x2f8
    faces: NP_x3i4


class ShapeActions(AnimatableActions):
    __slots__ = ()

    @Action.register()
    @classmethod
    def interpolate(
        cls: type[Self],
        dst: Shape,
        src_0: Shape,
        src_1: Shape
    ) -> Iterator[Animation]:
        yield ShapeInterpolateAnimation(dst, src_0, src_1)

    @Action.register()
    @classmethod
    def piecewise(
        cls: type[Self],
        dst: Shape,
        src: Shape,
        piecewiser: Piecewiser
    ) -> Iterator[Animation]:
        yield ShapePiecewiseAnimation(dst, src, piecewiser)


class Shape(ShapeActions, Animatable):
    __slots__ = ()

    def __init__(
        self: Self,
        # Should satisfy `all(counts > 0)` and `len(coordinates) == sum(counts)`.
        coordinates: NP_x2f8 | None = None,
        counts: NP_xi4 | None = None
    ) -> None:
        super().__init__()
        if coordinates is not None:
            self._coordinates_ = coordinates
        if counts is not None:
            self._counts_ = counts

    def __and__(
        self: Self,
        other: Self
    ) -> Self:
        return self.intersection(other)

    def __or__(
        self: Self,
        other: Self
    ) -> Self:
        return self.union(other)

    def __sub__(
        self: Self,
        other: Self
    ) -> Self:
        return self.difference(other)

    def __xor__(
        self: Self,
        other: Self
    ) -> Self:
        return self.xor(other)

    @Lazy.variable()
    @staticmethod
    def _coordinates_() -> NP_x2f8:
        return np.zeros((0, 2))

    @Lazy.variable()
    @staticmethod
    def _counts_() -> NP_xi4:
        return np.zeros((0,), dtype=np.int32)

    @Lazy.property()
    @staticmethod
    def _cumcounts_(
        counts: NP_xi4
    ) -> NP_xi4:
        return np.insert(np.cumsum(counts), 0, 0)

    @Lazy.property()
    @staticmethod
    def _graph_(
        coordinates: NP_x2f8,
        cumcounts: NP_xi4
    ) -> Graph:
        edge_starts = np.arange(len(coordinates), dtype=np.int32)
        edge_stops = np.arange(len(coordinates), dtype=np.int32)
        edge_stops[cumcounts[:-1]] = np.roll(edge_stops[cumcounts[:-1]], 1)
        edge_stops = np.roll(edge_stops, -1)
        return Graph(
            positions=np.concatenate((coordinates, np.zeros((len(coordinates), 1))), axis=1),
            edges=np.column_stack((edge_starts, edge_stops))
        )

    @Lazy.property()
    @staticmethod
    def _triangulation_(
        coordinates: NP_x2f8,
        cumcounts: NP_xi4
    ) -> Triangulation:

        def iter_contour_nodes(
            poly_trees: list[pyclipr.PolyTreeD]
        ) -> Iterator[pyclipr.PolyTreeD]:
            # http://www.angusj.com/clipper2/Docs/Units/Clipper.Engine/Classes/PolyTreeD/_Body.htm
            for poly_tree in poly_trees:
                yield poly_tree
                for hole in poly_tree.children:
                    yield from iter_contour_nodes(hole.children)

        def get_contour_triangulation(
            contour: pyclipr.PolyTreeD
        ) -> tuple[NP_x2f8, NP_x3i4]:
            ring_coordinates_tuple: tuple[NP_x2f8, ...] = (
                contour.polygon,
                *(hole.polygon for hole in contour.children)
            )
            coordinates = np.concatenate(ring_coordinates_tuple)
            ring_ends = np.fromiter((
                len(ring_coordinates) for ring_coordinates in ring_coordinates_tuple
            ), dtype=np.uint32).cumsum()
            return (
                coordinates,
                mapbox_earcut.triangulate_float64(coordinates, ring_ends).reshape((-1, 3)).astype(np.int32)
            )

        clipper = pyclipr.Clipper()
        clipper.addPaths([
            coordinates[start:stop]
            for start, stop in itertools.pairwise(cumcounts)
        ], pyclipr.Subject)
        poly_tree_root = clipper.execute2(pyclipr.Union, pyclipr.EvenOdd)
        triangulation_coordinates, triangulation_faces = ShapeUtils.concatenate_triangulations(
            get_contour_triangulation(contour)
            for contour in iter_contour_nodes(poly_tree_root.children)
        )
        return Triangulation(
            coordinates=triangulation_coordinates,
            faces=triangulation_faces
        )

    def set(
        self: Self,
        coordinates: NP_x2f8,
        counts: NP_xi4
    ) -> Self:
        self._coordinates_ = coordinates
        self._counts_ = counts
        return self

    def as_paths(
        self: Self,
        paths: Iterator[NP_x2f8]
    ) -> Self:
        cls = type(self)
        return self.concatenate(tuple(
            cls().set(
                coordinates=path,
                counts=np.array((len(path),), dtype=np.int32)
            )
            for path in paths
            if len(path)
        ))

    def as_clipping(
        self: Self,
        *shape_path_type_pairs: tuple[Self, pyclipr.PathType],
        # http://www.angusj.com/clipper2/Docs/Units/Clipper/Types/ClipType.htm
        clip_type: pyclipr.ClipType,
        # http://www.angusj.com/clipper2/Docs/Units/Clipper/Types/FillRule.htm
        fill_type: pyclipr.FillType
    ) -> Self:
        clipper = pyclipr.Clipper()
        for shape, path_type in shape_path_type_pairs:
            clipper.addPaths([
                shape._coordinates_[start:stop]
                for start, stop in itertools.pairwise(shape._cumcounts_)
            ], path_type)
        path_list: list[NP_x2f8] = clipper.execute(clip_type, fill_type)
        return self.as_paths(iter(path_list))

    def split(
        self: Self,
        dsts: tuple[Self, ...],
        alphas: NP_xf8
    ) -> Self:
        for dst, (coordinates, counts) in zip(dsts, ShapeUtils.shape_split(self, alphas), strict=True):
            dst.set(
                coordinates=coordinates,
                counts=counts
            )
        return self

    def concatenate(
        self: Self,
        srcs: tuple[Self, ...]
    ) -> Self:
        coordinates, counts = ShapeUtils.shape_concatenate(srcs)
        self.set(
            coordinates=coordinates,
            counts=counts
        )
        return self

    def intersection(
        self: Self,
        other: Self
    ) -> Self:
        return type(self)().as_clipping(
            (self, pyclipr.Subject),
            (other, pyclipr.Clip),
            clip_type=pyclipr.Intersection,
            fill_type=pyclipr.NonZero
        )

    def union(
        self: Self,
        other: Self
    ) -> Self:
        return type(self)().as_clipping(
            (self, pyclipr.Subject),
            (other, pyclipr.Clip),
            clip_type=pyclipr.Union,
            fill_type=pyclipr.NonZero
        )

    def difference(
        self: Self,
        other: Self
    ) -> Self:
        return type(self)().as_clipping(
            (self, pyclipr.Subject),
            (other, pyclipr.Clip),
            clip_type=pyclipr.Difference,
            fill_type=pyclipr.NonZero
        )

    def xor(
        self: Self,
        other: Self
    ) -> Self:
        return type(self)().as_clipping(
            (self, pyclipr.Subject),
            (other, pyclipr.Clip),
            clip_type=pyclipr.Xor,
            fill_type=pyclipr.NonZero
        )

    def animate(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> DynamicShape[Self]:
        return DynamicShape(self, **kwargs)


class DynamicShape[ShapeT: Shape](ShapeActions, DynamicAnimatable[ShapeT]):
    __slots__ = ()


@attrs.frozen(kw_only=True)
class ShapeInterpolateInfo:
    coordinates_0: NP_x2f8
    coordinates_1: NP_x2f8
    counts: NP_xi4


class ShapeInterpolateAnimation[ShapeT: Shape](AnimatableInterpolateAnimation[ShapeT]):
    __slots__ = ()

    @Lazy.property()
    @staticmethod
    def _interpolate_info_(
        src_0: ShapeT,
        src_1: ShapeT
    ) -> ShapeInterpolateInfo:
        coordinates_0, coordinates_1, counts = ShapeUtils.shape_interpolate(
            shape_0=src_0,
            shape_1=src_1
        )
        return ShapeInterpolateInfo(
            coordinates_0=coordinates_0,
            coordinates_1=coordinates_1,
            counts=counts
        )

    def interpolate(
        self: Self,
        dst: ShapeT,
        alpha: float
    ) -> None:
        interpolate_info = self._interpolate_info_
        dst.set(
            coordinates=(1.0 - alpha) * interpolate_info.coordinates_0 + alpha * interpolate_info.coordinates_1,
            counts=interpolate_info.counts
        )


class ShapePiecewiseAnimation[ShapeT: Shape](AnimatablePiecewiseAnimation[ShapeT]):
    __slots__ = ()

    @classmethod
    def split(
        cls: type[Self],
        dsts: tuple[ShapeT, ...],
        src: ShapeT,
        alphas: NP_xf8
    ) -> None:
        src.split(dsts, alphas)

    @classmethod
    def concatenate(
        cls: type[Self],
        dst: ShapeT,
        srcs: tuple[ShapeT, ...]
    ) -> None:
        dst.concatenate(srcs)


class ShapeUtils(GraphUtils):
    __slots__ = ()

    @classmethod
    def concatenate_triangulations(
        cls: type[Self],
        triangulations: Iterator[tuple[NP_x2f8, NP_x3i4]]
    ) -> tuple[NP_x2f8, NP_x3i4]:
        triangulation_tuple: tuple[tuple[NP_x2f8, NP_x3i4], ...] = (
            (np.zeros((0, 2)), np.zeros((0, 3), dtype=np.int32)),
            *triangulations
        )
        offsets = np.roll(np.fromiter((
            len(coordinates) for coordinates, _ in triangulation_tuple
        ), dtype=np.int32).cumsum(), 1)
        return (
            np.concatenate(tuple(coordinates for coordinates, _ in triangulation_tuple)),
            np.concatenate(tuple(faces + offset for (_, faces), offset in zip(triangulation_tuple, offsets, strict=True)))
        )

    @classmethod
    def shape_split(
        cls: type[Self],
        shape: Shape,
        alphas: NP_xf8
    ) -> Iterator[tuple[NP_x2f8, NP_xi4]]:
        for positions, edges in cls.graph_split(shape._graph_, alphas):
            (coordinates,), counts = cls._unified_graphs_to_unified_shapes((positions,), edges)
            yield coordinates, counts

    @classmethod
    def shape_concatenate(
        cls: type[Self],
        shapes: tuple[Shape, ...]
    ) -> tuple[NP_x2f8, NP_xi4]:
        positions, edges = cls.graph_concatenate(tuple(shape._graph_ for shape in shapes))
        (coordinates,), counts = cls._unified_graphs_to_unified_shapes((positions,), edges)
        return coordinates, counts

    @classmethod
    def shape_interpolate(
        cls: type[Self],
        shape_0: Shape,
        shape_1: Shape
    ) -> tuple[NP_x2f8, NP_x2f8, NP_xi4]:
        graph_0 = shape_0._graph_
        graph_1 = shape_1._graph_
        positions_0, positions_1, edges = cls._unify_aligned_graph_pairs(
            *cls._iter_aligned_graph_pairs_for_shape_interpolation(
                positions_0=graph_0._positions_,
                edges_0=graph_0._edges_,
                cumlengths_0=graph_0._cumlengths_,
                cumcounts_0=shape_0._cumcounts_,
                positions_1=graph_1._positions_,
                edges_1=graph_1._edges_,
                cumlengths_1=graph_1._cumlengths_,
                cumcounts_1=shape_1._cumcounts_
            )
        )
        (coordinates_0, coordinates_1), counts = cls._unified_graphs_to_unified_shapes((positions_0, positions_1), edges)
        return coordinates_0, coordinates_1, counts

    @classmethod
    def _unified_graphs_to_unified_shapes(
        cls: type[Self],
        positions_tuple: tuple[NP_x3f8, ...],
        edges: NP_x2i4
    ) -> tuple[tuple[NP_x2f8, ...], NP_xi4]:
        if not len(edges):
            indices = np.zeros((0,), dtype=np.int32)
            counts = np.zeros((0,), dtype=np.int32)
        else:
            disjoints = np.array((0, *np.flatnonzero(edges[:-1, 1] - edges[1:, 0]) + 1, len(edges)))
            open_path_indices = np.flatnonzero(edges[disjoints[:-1], 0] - edges[disjoints[1:] - 1, 1])
            indices = np.insert(edges[:, 0], disjoints[open_path_indices + 1], edges[disjoints[open_path_indices + 1] - 1, 1])
            counts = np.diff(disjoints)
            counts[open_path_indices] += 1
        return (
            tuple(
                positions[indices, :2]
                for positions in positions_tuple
            ),
            counts
        )

    @classmethod
    def _iter_aligned_graph_pairs_for_shape_interpolation(
        cls: type[Self],
        positions_0: NP_x3f8,
        edges_0: NP_x2i4,
        cumlengths_0: NP_xf8,
        cumcounts_0: NP_xi4,
        positions_1: NP_x3f8,
        edges_1: NP_x2i4,
        cumlengths_1: NP_xf8,
        cumcounts_1: NP_xi4
    ) -> Iterator[tuple[tuple[NP_x3f8, NP_x2i4], tuple[NP_x3f8, NP_x2i4]]]:
        yield from cls._iter_aligned_graph_pairs_for_graph_interpolation(
            positions_0=positions_0,
            edges_0=edges_0,
            cumlengths_0=cumlengths_0,
            positions_1=positions_1,
            edges_1=edges_1,
            cumlengths_1=cumlengths_1
        )

        if not len(edges_0) or not len(edges_1):
            return

        knots_0 = cumlengths_0 * cumlengths_1[-1]
        knots_1 = cumlengths_1 * cumlengths_0[-1]
        inlay_positions_0, inlay_edges_0 = cls._get_inlay_graph(
            positions=positions_0,
            edges=edges_0,
            knots=knots_0,
            cumcounts=cumcounts_0,
            alphas=knots_1[cumcounts_1][1:-1],
            side="right"
        )
        inlay_positions_1, inlay_edges_1 = cls._get_inlay_graph(
            positions=positions_1,
            edges=edges_1,
            knots=knots_1,
            cumcounts=cumcounts_1,
            alphas=knots_0[cumcounts_0][1:-1],
            side="left"
        )
        yield (
            (inlay_positions_0, inlay_edges_0),
            cls._get_centroid_graph(positions_1, edges_1, inlay_edges_0)
        )
        yield (
            cls._get_centroid_graph(positions_0, edges_0, inlay_edges_1),
            (inlay_positions_1, inlay_edges_1)
        )

    @classmethod
    def _get_inlay_graph(
        cls: type[Self],
        positions: NP_x3f8,
        edges: NP_x2i4,
        knots: NP_xf8,
        cumcounts: NP_xi4,
        alphas: NP_xf8,
        side: Literal["left", "right"]
    ) -> tuple[NP_x3f8, NP_x2i4]:
        inlay_interpolated_positions, inlay_insertion_indices = cls._get_interpolated_samples(
            positions=positions,
            edges=edges,
            knots=knots,
            alphas=alphas,
            side=side
        )
        return cls._insert_samples(
            positions=positions,
            edges=np.column_stack((
                edges[cumcounts[:-1], 0],
                edges[cumcounts[1:] - 1, 1]
            )),
            interpolated_positions=inlay_interpolated_positions,
            insertion_indices=np.searchsorted(
                cumcounts[1:-1] - 1,
                inlay_insertion_indices,
                side=side
            ).astype(np.int32)
        )




from __future__ import annotations


import pathlib
from typing import (
    Iterator,
    Self,
    TypedDict
)

import attrs
import numpy as np
import svgelements as se
from scipy.interpolate import BSpline

from ..animatables.arrays.animatable_color import AnimatableColor
from ..animatables.shape import Shape
from ..constants.custom_typing import (
    NP_2f8,
    NP_x2f8
)
from .shape_mobjects.shape_mobject import ShapeMobject
from .image_mobject import ImageMobject
from .mobject_io import (
    MobjectIO,
    MobjectInput,
    MobjectJSON,
    MobjectOutput
)


@attrs.frozen(kw_only=True)
class SVGMobjectInput(MobjectInput):
    svg_path: pathlib.Path
    svg_text: str
    width: float | None
    height: float | None
    scale: float | None


@attrs.frozen(kw_only=True)
class SVGMobjectOutput(MobjectOutput):
    shape_mobjects: tuple[ShapeMobject, ...]


class ShapeMobjectJSON(TypedDict):
    coordinates: tuple[float, ...]  # flattened
    counts: tuple[int, ...]
    color: str
    opacity: float


class SVGMobjectJSON(MobjectJSON):
    shape_mobjects: tuple[ShapeMobjectJSON, ...]


class SVGMobjectIO(MobjectIO[SVGMobjectInput, SVGMobjectOutput, SVGMobjectJSON]):
    __slots__ = ()

    @property
    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "svg_mobject"

    @classmethod
    def generate(
        cls: type[Self],
        input_data: SVGMobjectInput,
        temp_path: pathlib.Path
    ) -> SVGMobjectOutput:
        return SVGMobjectOutput(
            shape_mobjects=cls._get_shape_mobjects_from_svg_path(
                svg_path=input_data.svg_path,
                width=input_data.width,
                height=input_data.height,
                scale=input_data.scale
            )
        )

    @classmethod
    def dump_json(
        cls: type[Self],
        output_data: SVGMobjectOutput
    ) -> SVGMobjectJSON:
        return SVGMobjectJSON(
            shape_mobjects=tuple(
                cls._shape_mobject_to_json(shape_mobject)
                for shape_mobject in output_data.shape_mobjects
            )
        )

    @classmethod
    def load_json(
        cls: type[Self],
        json_data: SVGMobjectJSON
    ) -> SVGMobjectOutput:
        return SVGMobjectOutput(
            shape_mobjects=tuple(
                cls._json_to_shape_mobject(shape_mobject_json)
                for shape_mobject_json in json_data["shape_mobjects"]
            )
        )

    @classmethod
    def _get_shape_mobjects_from_svg_path(
        cls: type[Self],
        svg_path: str | pathlib.Path,
        *,
        width: float | None = None,
        height: float | None = None,
        scale: float | None = None
    ) -> tuple[ShapeMobject, ...]:

        def perspective(
            origin_x: float,
            origin_y: float,
            radius_x: float,
            radius_y: float
        ) -> se.Matrix:
            # `(origin=(0.0, 0.0), radius=(1.0, 1.0))` ->
            # `(origin=(origin_x, origin_y), radius=(radius_x, radius_y))`
            return se.Matrix(
                radius_x,
                0.0,
                0.0,
                radius_y,
                origin_x,
                origin_y
            )

        def get_transform(
            bbox: tuple[float, float, float, float],
            width: float | None,
            height: float | None,
            scale: float | None
        ) -> se.Matrix:

            min_x, min_y, max_x, max_y = bbox
            origin_x = (min_x + max_x) / 2.0
            origin_y = (min_y + max_y) / 2.0
            radius_x = (max_x - min_x) / 2.0
            radius_y = (max_y - min_y) / 2.0
            transform = ~perspective(
                origin_x=origin_x,
                origin_y=origin_y,
                radius_x=radius_x,
                radius_y=radius_y
            )
            scale_x, scale_y = ImageMobject._get_scale_vector(
                original_width=radius_x * 2.0,
                original_height=radius_y * 2.0,
                specified_width=width,
                specified_height=height,
                specified_scale=scale
            )
            transform *= perspective(
                origin_x=0.0,
                origin_y=0.0,
                radius_x=scale_x * radius_x,
                radius_y=-scale_y * radius_y  # Flip y.
            )
            return transform

        def iter_paths_from_se_shape(
            se_shape: se.Shape
        ) -> Iterator[NP_x2f8]:
            se_path = se.Path(se_shape.segments(transformed=True))
            se_path.approximate_arcs_with_cubics()
            coordinates_list: list[NP_2f8] = []
            is_ring: bool = False
            coordinates_dtype = np.dtype((np.float64, (2,)))
            for segment in se_path.segments(transformed=True):
                match segment:
                    case se.Move(end=end):
                        if is_ring:
                            yield np.fromiter(coordinates_list, dtype=coordinates_dtype)
                        coordinates_list = [np.array(end)]
                        is_ring = False
                    case se.Close():
                        is_ring = True
                    case se.Line(end=end):
                        coordinates_list.append(np.array(end))
                    case se.QuadraticBezier() | se.CubicBezier():
                        # Approximate the bezier curve with a polyline.
                        control_positions = np.array(segment)
                        curve = BSpline(
                            t=np.repeat(np.array((0.0, 1.0)), len(control_positions)),
                            c=control_positions,
                            k=len(control_positions) - 1
                        )
                        coordinates_list.extend(curve(np.linspace(0.0, 1.0, 9)[1:]))
                    case _:
                        raise ValueError(f"Cannot handle path segment type: {type(segment)}")
            if is_ring:
                yield np.fromiter(coordinates_list, dtype=coordinates_dtype)

        def iter_shape_mobjects_from_svg(
            svg: se.SVG
        ) -> Iterator[ShapeMobject]:
            bbox: tuple[float, float, float, float] | None = svg.bbox()
            if bbox is None:
                return

            # Handle transform before constructing `Shape`s,
            # so that the center of the entire shape falls on the origin.
            transform = get_transform(
                bbox=bbox,
                width=width,
                height=height,
                scale=scale
            )

            for se_shape in svg.elements():
                if not isinstance(se_shape, se.Shape):
                    continue
                shape = Shape().as_paths(iter_paths_from_se_shape(se_shape * transform))
                if not len(shape._coordinates_):
                    # Filter out empty shapes.
                    continue
                style_dict = {}
                if se_shape.fill is not None:
                    if (color := se_shape.fill.hexrgb) is not None:
                        style_dict["color"] = color
                    if (opacity := se_shape.fill.opacity) is not None:
                        style_dict["opacity"] = opacity
                yield ShapeMobject(shape).set(**style_dict)

        svg: se.SVG = se.SVG.parse(svg_path)
        return tuple(iter_shape_mobjects_from_svg(svg))

    @classmethod
    def _shape_mobject_to_json(
        cls: type[Self],
        shape_mobject: ShapeMobject
    ) -> ShapeMobjectJSON:
        shape = shape_mobject._shape_
        return ShapeMobjectJSON(
            coordinates=tuple(round(float(value), 6) for value in shape._coordinates_.flatten()),
            counts=tuple(int(value) for value in shape._counts_),
            color=AnimatableColor._array_to_hex(shape_mobject._color_._array_),
            opacity=round(float(shape_mobject._opacity_._array_), 6)
        )

    @classmethod
    def _json_to_shape_mobject(
        cls: type[Self],
        shape_mobject_json: ShapeMobjectJSON
    ) -> ShapeMobject:
        coordinates = shape_mobject_json["coordinates"]
        counts = shape_mobject_json["counts"]
        color = shape_mobject_json["color"]
        opacity = shape_mobject_json["opacity"]

        return ShapeMobject(Shape(
            coordinates=np.fromiter(coordinates, dtype=np.float64).reshape(-1, 2),
            counts=np.fromiter(counts, dtype=np.int32)
        )).set(
            color=color,
            opacity=opacity
        )


class SVGMobject(ShapeMobject):
    __slots__ = ("_shape_mobjects",)

    def __init__(
        self: Self,
        svg_path: str | pathlib.Path,
        *,
        width: float | None = None,
        height: float | None = None,
        scale: float | None = None
    ) -> None:
        super().__init__()
        svg_path = pathlib.Path(svg_path)
        output_data = SVGMobjectIO.get(SVGMobjectInput(
            svg_path=svg_path,
            svg_text=svg_path.read_text(encoding="utf-8"),
            width=width,
            height=height,
            scale=scale
        ))

        shape_mobjects = output_data.shape_mobjects
        self._shape_mobjects: tuple[ShapeMobject, ...] = shape_mobjects
        self.add(*shape_mobjects)



from __future__ import annotations


from typing import Self

import numpy as np

from .polygon import Polygon


class Square(Polygon):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        super().__init__(coordinates=np.array((
            (1.0, 1.0),
            (-1.0, 1.0),
            (-1.0, -1.0),
            (1.0, -1.0)
        )))



from __future__ import annotations


from typing import Self

from ...constants.custom_typing import RateType
from ..timeline import Timeline
from .lagged import Lagged


class Parallel(Timeline):
    __slots__ = (
        "_timeline_items",
        "_rate"
    )

    def __init__(
        self: Self,
        *timelines: Timeline,
        rate: RateType | None = None,
        lag_time: float = 0.0,
        lag_ratio: float = 0.0
    ) -> None:
        accumulated_lag_time = 0.0
        timeline_items: list[tuple[Timeline, float]] = []
        for timeline in timelines:
            timeline_items.append((timeline, accumulated_lag_time))
            accumulated_lag_time += lag_time + lag_ratio * timeline._run_alpha
        super().__init__(
            run_alpha=max((
                timeline_lag_time + timeline._run_alpha
                for timeline, timeline_lag_time in timeline_items
            ), default=0.0)
        )
        self._timeline_items: tuple[tuple[Timeline, float], ...] = tuple(timeline_items)
        self._rate: RateType | None = rate

    async def construct(
        self: Self
    ) -> None:
        timeline_items = self._timeline_items
        rate = self._rate
        for timeline, timeline_lag_time in timeline_items:
            self.prepare(Lagged(timeline, lag_time=timeline_lag_time), rate=rate)
        await self.wait_until(lambda: all(
            timeline.terminated() for timeline, _ in timeline_items
        ))



from __future__ import annotations


import json
import pathlib
import subprocess
from typing import (
    Iterator,
    Self,
    Unpack
)

import attrs

from ...toplevel.toplevel import Toplevel
from .pango_string_mobject import (
    PangoAttributes,
    PangoStringMobjectIO,
    PangoStringMobjectInput,
    PangoStringMobjectKwargs
)
from .string_mobject import (
    Span,
    StringMobject
)


@attrs.frozen(kw_only=True)
class CodeInput(PangoStringMobjectInput):
    font: str = attrs.field(factory=lambda: Toplevel._get_config().code_font)
    language_suffix: str = attrs.field(factory=lambda: Toplevel._get_config().code_language_suffix)


class CodeKwargs(PangoStringMobjectKwargs, total=False):
    language_suffix: str


class CodeIO[CodeInputT: CodeInput](PangoStringMobjectIO[CodeInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "code"

    @classmethod
    def _iter_local_span_attributes(
        cls: type[Self],
        input_data: CodeInputT,
        temp_path: pathlib.Path
    ) -> Iterator[tuple[Span, PangoAttributes]]:
        language_suffix = input_data.language_suffix
        try:
            code_path = temp_path.with_suffix(language_suffix)
            code_path.write_text(input_data.string, encoding="utf-8")

            # First open the file, then launch the command.
            # We separate these two steps as file loading is asynchronous,
            # and operations on `view` has no effect while loading.
            if subprocess.run((
                "subl",
                code_path,
                "--background"   # Don't activate the application.
            ), stdout=subprocess.DEVNULL).returncode or subprocess.run((
                "subl",
                "--background",
                "--command", "export_highlight"
            ), stdout=subprocess.DEVNULL).returncode:
                raise OSError("CodeIO: Failed to execute subl command")

            json_str = temp_path.with_suffix(".json").read_text(encoding="utf-8")
            for token in json.loads(json_str):
                # See `https://www.sublimetext.com/docs/api_reference.html#sublime.View.style_for_scope`.
                style = token["style"]
                attributes = PangoAttributes()
                if (foreground := style.get("foreground")):
                    attributes["fgcolor"] = foreground
                if (background := style.get("background")):
                    attributes["bgcolor"] = background
                if style.get("bold"):
                    attributes["font_weight"] = "bold"
                if style.get("italic"):
                    attributes["font_style"] = "italic"
                yield Span(token["begin"], token["end"]), attributes
        finally:
            for suffix in (language_suffix, ".json"):
                temp_path.with_suffix(suffix).unlink(missing_ok=True)

        yield from super()._iter_local_span_attributes(input_data, temp_path)


class Code(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[CodeKwargs]
    ) -> None:
        super().__init__(CodeIO.get(CodeInput(string=string, **kwargs)))



from __future__ import annotations


import math
from typing import Self

import numpy as np
from scipy.spatial.transform import Rotation

from ...constants.custom_typing import (
    NP_x3f8,
    NP_xxi4
)
from .polygon import Polygon
from .shape_mobject import ShapeMobject


class Polyhedron(ShapeMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        positions: NP_x3f8,
        faces: NP_xxi4
    ) -> None:
        super().__init__()
        self.add(*(
            type(self)._get_transformed_face(positions[face])
            for face in faces
        ))

    @classmethod
    def _get_transformed_face(
        cls: type[Self],
        positions: NP_x3f8
    ) -> Polygon:
        assert len(positions) >= 3
        # We first choose three sample positions that define the plane.
        # Instead of choosing `positions[:3]`, we choose `positions[:2]` and the geometric centroid,
        # in order to reduce the chance that they happen to be colinear.
        sample_0 = np.average(positions, axis=0)
        sample_1 = positions[0]
        sample_2 = positions[1]
        x_axis = (x_direction := sample_1 - sample_0) / np.linalg.norm(x_direction)
        z_axis = (z_direction := np.cross(x_axis, sample_2 - sample_0)) / np.linalg.norm(z_direction)
        y_axis = np.cross(z_axis, x_axis)
        rotation_matrix = np.vstack((x_axis, y_axis, z_axis)).T
        rotation_vector = Rotation.from_matrix(rotation_matrix).as_rotvec()

        transformed = (np.linalg.inv(rotation_matrix) @ (positions - sample_0).T).T
        transformed_xy = transformed[:, :2]
        transformed_z = transformed[:, 2]
        assert np.isclose(transformed_z, 0.0).all(), "Positions are not coplanar"
        return Polygon(transformed_xy).rotate_about_origin(rotation_vector).shift(sample_0)


# The five Platonic solids, ported from manim community `/manim/mobject/three_d/polyhedra.py`
# All these polyhedra have all positions sitting on the unit sphere.
class Tetrahedron(Polyhedron):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        a = 1.0 / math.sqrt(3.0)
        super().__init__(
            positions=np.array((
                ( a,  a,  a),
                ( a, -a, -a),
                (-a,  a, -a),
                (-a, -a,  a)
            )),
            faces=np.array((
                (0, 1, 2),
                (3, 0, 2),
                (1, 0, 3),
                (2, 1, 3)
            ))
        )


class Cube(Polyhedron):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        a = 1.0 / math.sqrt(3.0)
        super().__init__(
            positions=np.array((
                ( a,  a,  a),
                ( a,  a, -a),
                ( a, -a,  a),
                ( a, -a, -a),
                (-a,  a,  a),
                (-a,  a, -a),
                (-a, -a,  a),
                (-a, -a, -a)
            )),
            faces=np.array((
                (0, 2, 3, 1),
                (4, 5, 7, 6),
                (0, 1, 5, 4),
                (2, 6, 7, 3),
                (0, 4, 6, 2),
                (1, 3, 7, 5)
            ))
        )


class Octahedron(Polyhedron):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        a = 1.0
        super().__init__(
            positions=np.array((
                (  a, 0.0, 0.0),
                ( -a, 0.0, 0.0),
                (0.0,   a, 0.0),
                (0.0,  -a, 0.0),
                (0.0, 0.0,   a),
                (0.0, 0.0,  -a)
            )),
            faces=np.array((
                (0, 2, 4),
                (2, 1, 4),
                (1, 3, 4),
                (3, 0, 4),
                (0, 3, 5),
                (3, 1, 5),
                (1, 2, 5),
                (2, 0, 5)
            ))
        )


class Dodecahedron(Polyhedron):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        a = (math.sqrt(5.0) + 1.0) / (2.0 * math.sqrt(3.0))
        b = 1.0 / math.sqrt(3.0)
        c = (math.sqrt(5.0) - 1.0) / (2.0 * math.sqrt(3.0))
        super().__init__(
            positions=np.array((
                (  b,   b,   b),
                (  b,   b,  -b),
                (  b,  -b,   b),
                (  b,  -b,  -b),
                ( -b,   b,   b),
                ( -b,   b,  -b),
                ( -b,  -b,   b),
                ( -b,  -b,  -b),
                (0.0,   a,   c),
                (0.0,   a,  -c),
                (0.0,  -a,   c),
                (0.0,  -a,  -c),
                (  c, 0.0,   a),
                ( -c, 0.0,   a),
                (  c, 0.0,  -a),
                ( -c, 0.0,  -a),
                (  a,   c, 0.0),
                (  a,  -c, 0.0),
                ( -a,   c, 0.0),
                ( -a,  -c, 0.0)
            )),
            faces=np.array((
                ( 8,  0, 16,  1,  9),
                ( 9,  5, 18,  4,  8),
                (10,  6, 19,  7, 11),
                (11,  3, 17,  2, 10),
                (12,  0,  8,  4, 13),
                (13,  6, 10,  2, 12),
                (14,  3, 11,  7, 15),
                (15,  5,  9,  1, 14),
                (16,  0, 12,  2, 17),
                (17,  3, 14,  1, 16),
                (18,  5, 15,  7, 19),
                (19,  6, 13,  4, 18)
            ))
        )


class Icosahedron(Polyhedron):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        a = math.sqrt((1.0 + 1.0 / math.sqrt(5.0)) / 2.0)
        b = math.sqrt((1.0 - 1.0 / math.sqrt(5.0)) / 2.0)
        super().__init__(
            positions=np.array((
                (0.0,   a,   b),
                (0.0,   a,  -b),
                (0.0,  -a,   b),
                (0.0,  -a,  -b),
                (  b, 0.0,   a),
                ( -b, 0.0,   a),
                (  b, 0.0,  -a),
                ( -b, 0.0,  -a),
                (  a,   b, 0.0),
                (  a,  -b, 0.0),
                ( -a,   b, 0.0),
                ( -a,  -b, 0.0)
            )),
            faces=np.array((
                ( 0,  8,  1),
                ( 1, 10,  0),
                ( 2, 11,  3),
                ( 3,  9,  2),
                ( 4,  0,  5),
                ( 5,  2,  4),
                ( 6,  3,  7),
                ( 7,  1,  6),
                ( 8,  4,  9),
                ( 9,  6,  8),
                (10,  7, 11),
                (11,  5, 10),
                ( 8,  0,  4),
                ( 0, 10,  5),
                (11,  2,  5),
                ( 2,  9,  4),
                ( 9,  3,  6),
                ( 3, 11,  7),
                (10,  1,  7),
                ( 1,  8,  6)
            ))
        )



from __future__ import annotations


from typing import Self

import numpy as np

from ...animatables.shape import Shape
from ...constants.custom_typing import NP_x2f8
from .shape_mobject import ShapeMobject


class Polygon(ShapeMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        coordinates: NP_x2f8
    ) -> None:
        super().__init__(Shape(
            coordinates=coordinates,
            counts=np.array((len(coordinates),))
        ))



from __future__ import annotations


from typing import Self

import attrs
import moderngl

from ...toplevel.toplevel import Toplevel
from ..mgl_enums import (
    BlendEquation,
    BlendFunc,
    ContextFlag
)
from ..vertex_array import VertexArray


@attrs.frozen(kw_only=True)
class AttachmentInfo:
    components: int
    dtype: str
    src_blend_func: BlendFunc
    dst_blend_func: BlendFunc
    blend_equation: BlendEquation


class Framebuffer:
    __slots__ = (
        "_named_attachments",
        "_framebuffer",
        "_blendings",
        "_flag"
    )

    def __init__(
        self: Self,
        attachment_info_dict: dict[str, AttachmentInfo],
        samples: int,
        flag: ContextFlag
    ) -> None:
        super().__init__()
        size = Toplevel._get_config().pixel_size
        attachment_items = tuple(
            (
                name,
                Toplevel._get_context().texture(
                    size=size,
                    components=attachment_info.components,
                    samples=samples,
                    dtype=attachment_info.dtype
                ),
                (attachment_info.src_blend_func, attachment_info.dst_blend_func, attachment_info.blend_equation)
            )
            for name, attachment_info in attachment_info_dict.items()
        )

        self._named_attachments: dict[str, moderngl.Texture] = {
            name: attachment for name, attachment, _ in attachment_items
        }
        self._framebuffer: moderngl.Framebuffer = Toplevel._get_context().framebuffer(color_attachments=tuple(
            attachment for _, attachment, _ in attachment_items
        ))
        self._blendings: tuple[tuple[BlendFunc, BlendFunc, BlendEquation], ...] = tuple(
            blending for _, _, blending in attachment_items
        )
        self._flag: ContextFlag = flag

    def clear(
        self: Self,
        color: tuple[float, float, float, float] | None = None
    ) -> None:
        self._framebuffer.clear(color=color)

    def render(
        self: Self,
        vertex_array: VertexArray
    ) -> None:
        if (vertex_array_info := vertex_array._vertex_array_info_) is None:
            return None

        uniform_block_buffer_dict = {
            uniform_block_buffer._name_: uniform_block_buffer
            for uniform_block_buffer in vertex_array._uniform_block_buffers_
        }

        with Toplevel._get_context().scope(
            framebuffer=self._framebuffer,
            textures=vertex_array_info.texture_bindings,
            uniform_buffers=tuple(
                (uniform_block_buffer_dict[name]._buffer_, binding)
                for name, binding in vertex_array_info.uniform_block_bindings
            )
        ):
            Toplevel._get_context().set_blendings(self._blendings)
            Toplevel._get_context().set_flag(self._flag)
            vertex_array_info.vertex_array.render()

    def copy_from(
        self: Self,
        framebuffer: Framebuffer
    ) -> None:
        Toplevel._get_context().copy_framebuffer(
            dst=self._framebuffer,
            src=framebuffer._framebuffer
        )



from __future__ import annotations


from enum import (
    Enum,
    Flag
)

import moderngl


class ContextFlag(Flag):
    NOTHING = moderngl.NOTHING
    BLEND = moderngl.BLEND
    DEPTH_TEST = moderngl.DEPTH_TEST
    CULL_FACE = moderngl.CULL_FACE
    RASTERIZER_DISCARD = moderngl.RASTERIZER_DISCARD
    PROGRAM_POINT_SIZE = moderngl.PROGRAM_POINT_SIZE


class PrimitiveMode(Enum):
    POINTS = moderngl.POINTS
    LINES = moderngl.LINES
    LINE_LOOP = moderngl.LINE_LOOP
    LINE_STRIP = moderngl.LINE_STRIP
    TRIANGLES = moderngl.TRIANGLES
    TRIANGLE_STRIP = moderngl.TRIANGLE_STRIP
    TRIANGLE_FAN = moderngl.TRIANGLE_FAN
    LINES_ADJACENCY = moderngl.LINES_ADJACENCY
    LINE_STRIP_ADJACENCY = moderngl.LINE_STRIP_ADJACENCY
    TRIANGLES_ADJACENCY = moderngl.TRIANGLES_ADJACENCY
    TRIANGLE_STRIP_ADJACENCY = moderngl.TRIANGLE_STRIP_ADJACENCY
    PATCHES = moderngl.PATCHES


class BlendFunc(Enum):
    ZERO = moderngl.ZERO
    ONE = moderngl.ONE
    SRC_COLOR = moderngl.SRC_COLOR
    ONE_MINUS_SRC_COLOR = moderngl.ONE_MINUS_SRC_COLOR
    SRC_ALPHA = moderngl.SRC_ALPHA
    ONE_MINUS_SRC_ALPHA = moderngl.ONE_MINUS_SRC_ALPHA
    DST_ALPHA = moderngl.DST_ALPHA
    ONE_MINUS_DST_ALPHA = moderngl.ONE_MINUS_DST_ALPHA
    DST_COLOR = moderngl.DST_COLOR
    ONE_MINUS_DST_COLOR = moderngl.ONE_MINUS_DST_COLOR


class BlendEquation(Enum):
    FUNC_ADD = moderngl.FUNC_ADD
    FUNC_SUBTRACT = moderngl.FUNC_SUBTRACT
    FUNC_REVERSE_SUBTRACT = moderngl.FUNC_REVERSE_SUBTRACT
    MIN = moderngl.MIN
    MAX = moderngl.MAX



from __future__ import annotations


import re
from typing import (
    Iterator,
    Self,
    Unpack
)

import attrs

from .pango_string_mobject import (
    PangoAttributes,
    PangoStringMobjectIO,
    PangoStringMobjectInput,
    PangoStringMobjectKwargs
)
from .string_mobject import (
    BalancedCommandInfo,
    CommandInfo,
    StandaloneCommandInfo,
    StringMobject
)


@attrs.frozen(kw_only=True)
class MarkupInput(PangoStringMobjectInput):
    pass


class MarkupKwargs(PangoStringMobjectKwargs, total=False):
    pass


class MarkupIO[MarkupInputT: MarkupInput](PangoStringMobjectIO[MarkupInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "markup"

    @classmethod
    def _iter_command_infos(
        cls: type[Self],
        string: str
    ) -> Iterator[CommandInfo[PangoAttributes]]:
        pattern = re.compile(r"""
            (?P<tag>
                <
                (?P<close_slash>/)?
                (?P<tag_name>\w+)\s*
                (?P<attr_list>(?:\w+\s*\=\s*(?P<quot>["']).*?(?P=quot)\s*)*)
                (?P<elision_slash>/)?
                >
            )
            |(?P<passthrough>
                <\?.*?\?>|<!--.*?-->|<!\[CDATA\[.*?\]\]>|<!DOCTYPE.*?>
            )
            |(?P<entity>&(?P<unicode>\#(?P<hex>x)?)?(?P<content>.*?);)
            |(?P<char>[>"'])
        """, flags=re.VERBOSE | re.DOTALL)
        attributes_pattern = re.compile(r"""
            (?P<attr_name>\w+)
            \s*\=\s*
            (?P<quot>["'])(?P<attr_val>.*?)(?P=quot)
        """, flags=re.VERBOSE | re.DOTALL)
        open_stack: list[re.Match[str]] = []
        for match in pattern.finditer(string):
            if match.group("tag"):
                if match.group("elision_slash"):
                    yield StandaloneCommandInfo(match)
                elif not match.group("close_slash"):
                    open_stack.append(match)
                else:
                    open_match_obj = open_stack.pop()
                    attributes = PangoAttributes({
                        attributes_match_obj.group("attr_name"): attributes_match_obj.group("attr_val")
                        for attributes_match_obj in attributes_pattern.finditer(open_match_obj.group("attr_list"))
                    }) if (tag_name := open_match_obj.group("tag_name")) == "span" else cls._MARKUP_TAGS[tag_name]
                    yield BalancedCommandInfo(
                        attributes=attributes,
                        isolated=False,
                        open_match_obj=open_match_obj,
                        close_match_obj=match,
                        open_replacement="",
                        close_replacement=""
                    )
            elif match.group("char"):
                yield StandaloneCommandInfo(match, replacement=cls._markup_escape(match.group("char")))
            else:
                yield StandaloneCommandInfo(match)
        assert not open_stack


class Markup(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[MarkupKwargs]
    ) -> None:
        super().__init__(MarkupIO.get(MarkupInput(string=string, **kwargs)))



from __future__ import annotations


import weakref
from typing import (
    TYPE_CHECKING,
    Hashable,
    Iterator,
    Self
)

if TYPE_CHECKING:
    from .lazy_descriptor import (
        LazyDescriptor,
        Memoized
    )


class LazySlot[T, DataT]:
    __slots__ = (
        "__weakref__",
        "_descriptor_ref",
        "_elements",
        "_parameter_key",
        "_associated_slots"
    )

    def __init__(
        self: Self,
        descriptor: LazyDescriptor[T, DataT]
    ) -> None:
        super().__init__()
        self._descriptor_ref: weakref.ref[LazyDescriptor[T, DataT]] = weakref.ref(descriptor)
        self._elements: tuple[Memoized[T], ...] | None = None
        self._parameter_key: Memoized[Hashable] | None = None
        self._associated_slots: weakref.WeakSet[LazySlot] = weakref.WeakSet()

    def get_descriptor(
        self: Self
    ) -> LazyDescriptor[T, DataT]:
        assert (descriptor := self._descriptor_ref()) is not None
        return descriptor

    def get(
        self: Self
    ) -> tuple[Memoized[T], ...] | None:
        return self._elements

    def set(
        self: Self,
        elements: tuple[Memoized[T], ...],
        parameter_key: Memoized[Hashable] | None,
        associated_slots: set[LazySlot]
    ) -> None:
        self._elements = elements
        self._parameter_key = parameter_key
        assert not self._associated_slots
        self._associated_slots.update(associated_slots)
        for slot in associated_slots:
            slot._associated_slots.add(self)

    def expire(
        self: Self
    ) -> None:
        self._elements = None
        for slot in self._associated_slots:
            slot._associated_slots.remove(self)
        self._associated_slots.clear()

    def iter_associated_slots(
        self: Self
    ) -> Iterator[LazySlot]:
        return iter(set(self._associated_slots))



from __future__ import annotations


import itertools
import time
from typing import (
    Iterator,
    Self
)

from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


class Timer(ToplevelResource):
    __slots__ = (
        "_start_timestamp",
        "_current_timestamp",
        "_fps_update_timestamp",
        "_recorded_fps",
        "_next_fps"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        timestamp = time.perf_counter()
        self._start_timestamp: float = timestamp
        self._current_timestamp: float = timestamp
        self._fps_update_timestamp: float = timestamp
        self._recorded_fps: int = 0
        self._next_fps: int = 0

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        Toplevel._timer = self
        yield
        Toplevel._timer = None

    def frame_clock(
        self: Self
    ) -> Iterator[float]:
        spf = 1.0 / Toplevel._get_config().fps
        # Integer-based counter for higher precision.
        for frame_index in itertools.count():
            timestamp = time.perf_counter()
            self._current_timestamp = timestamp
            if timestamp - self._fps_update_timestamp >= 1.0:
                self._fps_update_timestamp = timestamp
                self._recorded_fps = self._next_fps
                self._next_fps = 0
            self._next_fps += 1
            yield frame_index * spf
            if Toplevel._get_renderer()._livestreamer.is_livestreaming and (sleep_time := spf - (
                time.perf_counter() - timestamp
            )) > 0.0:
                time.sleep(sleep_time)



from __future__ import annotations


from typing import Self

from ...constants.custom_typing import RateType
from ..timeline import Timeline
from .parallel import Parallel


class Series(Parallel):
    __slots__ = ()

    def __init__(
        self: Self,
        *timelines: Timeline,
        rate: RateType | None = None,
        lag_time: float = 0.0,
        lag_ratio: float = 1.0
    ) -> None:
        super().__init__(
            *timelines,
            rate=rate,
            lag_time=lag_time,
            lag_ratio=lag_ratio
        )



from __future__ import annotations


from typing import Self

import numpy as np

from ...constants.constants import (
    PI,
    TAU
)
from ...constants.custom_typing import (
    NP_x2f8,
    NP_x3f8
)
from .parametric_surface import ParametricSurface


class Sphere(ParametricSurface):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:

        def func(
            samples: NP_x2f8
        ) -> NP_x3f8:
            theta = samples[:, 0]
            phi = samples[:, 1]
            return np.column_stack((np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)))

        super().__init__(
            func=func,
            normal_func=func,
            u_range=(0.0, TAU),
            v_range=(0.0, PI),
            resolution=(32, 16)
        )



from __future__ import annotations


from typing import Self

from ...constants.custom_typing import RateType
from ..timeline import Timeline


class Lagged(Timeline):
    __slots__ = (
        "_timeline",
        "_rate",
        "_lag_time"
    )

    def __init__(
        self: Self,
        timeline: Timeline,
        *,
        rate: RateType | None = None,
        lag_time: float = 0.0
    ) -> None:
        super().__init__(
            run_alpha=lag_time + timeline._run_alpha
        )
        self._timeline: Timeline = timeline
        self._rate: RateType | None = rate
        self._lag_time: float = lag_time

    async def construct(
        self: Self
    ) -> None:
        await self.wait(self._lag_time)
        await self.play(self._timeline, rate=self._rate)



from __future__ import annotations


import weakref
from typing import (
    Self,
    TypedDict,
    Unpack
)

from ...constants.custom_typing import RateType
from ...constants.rates import Rates
from ...lazy.lazy_object import LazyObject
from ...timelines.timeline import Timeline


class AnimateKwargs(TypedDict, total=False):
    rate: RateType
    rewind: bool
    infinite: bool


class Animation(LazyObject):
    __slots__ = ()

    def build(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> AnimationsTimeline:
        timeline = AnimationsTimeline(**kwargs)
        timeline._animations.append(self)
        return timeline

    def update(
        self: Self,
        alpha: float
    ) -> None:
        pass


class BodyAnimationsTimeline(Timeline):
    __slots__ = ("_animations_timeline_ref",)

    def __init__(
        self: Self,
        animations_timeline: AnimationsTimeline
    ) -> None:
        super().__init__(run_alpha=animations_timeline._run_alpha)
        self._animations_timeline_ref: weakref.ref[AnimationsTimeline] = weakref.ref(animations_timeline)

    def _animation_update(
        self: Self,
        time: float
    ) -> None:
        assert (animations_timeline := self._animations_timeline_ref()) is not None
        animations_timeline.update(time)

    async def construct(
        self: Self
    ) -> None:
        await self.wait(self._run_alpha)


class AnimationsTimeline(Timeline):
    __slots__ = (
        "_rate",
        "_animations"
    )

    def __init__(
        self: Self,
        rate: RateType = Rates.linear(),
        rewind: bool = False,
        infinite: bool = False
    ) -> None:
        super().__init__(run_alpha=float("inf") if infinite else 1.0)
        if rewind:
            rate = Rates.compose(rate, Rates.rewind())
        self._rate: RateType = rate
        self._animations: list[Animation] = []

    def update(
        self: Self,
        time: float
    ) -> None:
        alpha = self._rate(time)
        for animation in self._animations:
            animation.update(alpha)

    async def construct(
        self: Self
    ) -> None:
        self.update(0.0)
        await self.play(BodyAnimationsTimeline(self))
        self.update(1.0)



from __future__ import annotations


from typing import Self

import numpy as np

from ..constants.custom_typing import (
    NP_x2f8,
    NP_x3f8,
    NP_x3i4
)
from ..lazy.lazy import Lazy
from .animatable.animatable import Animatable


class Mesh(Animatable):
    __slots__ = ()

    def __init__(
        self: Self,
        positions: NP_x3f8 | None = None,
        normals: NP_x3f8 | None = None,
        uvs: NP_x2f8 | None = None,
        faces: NP_x3i4 | None = None
    ) -> None:
        super().__init__()
        if positions is not None:
            self._positions_ = positions
        if normals is not None:
            self._normals_ = normals
        if uvs is not None:
            self._uvs_ = uvs
        if faces is not None:
            self._faces_ = faces

    @Lazy.variable()
    @staticmethod
    def _positions_() -> NP_x3f8:
        return np.zeros((0, 3))

    @Lazy.variable()
    @staticmethod
    def _normals_() -> NP_x3f8:
        return np.zeros((0, 3))

    @Lazy.variable()
    @staticmethod
    def _uvs_() -> NP_x2f8:
        return np.zeros((0, 2))

    @Lazy.variable()
    @staticmethod
    def _faces_() -> NP_x3i4:
        return np.zeros((0, 3), dtype=np.int32)



from __future__ import annotations


from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Concatenate,
    Iterator,
    Self,
    overload
)

import attrs

from ...lazy.lazy_descriptor import LazyDescriptor

if TYPE_CHECKING:
    from .animatable import (
        Animatable,
        Animation,
        DynamicAnimatable
    )


@attrs.frozen(kw_only=True)
class DescriptorParameters:
    pass


@attrs.frozen(kw_only=True)
class ConverterDescriptorParameters(DescriptorParameters):
    converter: Callable[[Any], Animatable]


class Actions:
    __slots__ = ()

    def __init_subclass__(
        cls: type[Self]
    ) -> None:
        super().__init_subclass__()
        for action in cls.__dict__.values():
            if not isinstance(action, Action):
                continue
            action._actions_cls = cls


class Action[ActionsT: Actions, AnimatableT: Animatable, **P]:
    __slots__ = (
        "_method",
        "_actions_cls"
    )

    def __init__(
        self: Self,
        method: Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]
    ) -> None:
        super().__init__()
        self._method: Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]] = method
        self._actions_cls: type[ActionsT] = NotImplemented

    @overload
    def __get__[InstanceT: Animatable](
        self: Self,
        instance: InstanceT,
        owner: type[InstanceT] | None = None
    ) -> Callable[P, InstanceT]: ...

    @overload
    def __get__[InstanceT: DynamicAnimatable[Animatable]](
        self: Self,
        instance: InstanceT,
        owner: type[InstanceT] | None = None
    ) -> Callable[P, InstanceT]: ...

    @overload
    def __get__(
        self: Self,
        instance: None,
        owner: type[Actions] | None = None
    ) -> Self: ...

    def __get__(
        self: Self,
        instance: Actions | None,
        owner: type[Actions] | None = None
    ) -> Any:

        def get_bound_method(
            action: Self,
            instance: Any
        ) -> Any:

            def result(
                *args: P.args,
                **kwargs: P.kwargs
            ) -> Any:
                for animation in action.iter_animations(instance, *args, **kwargs):
                    animation.update(1.0)
                return instance

            return result

        def get_dynamic_bound_method(
            action: Self,
            instance: Any
        ) -> Any:

            def result(
                *args: P.args,
                **kwargs: P.kwargs
            ) -> Any:
                instance._animations.extend(action.iter_animations(instance._dst, *args, **kwargs))
                return instance

            return result

        from .animatable import (
            Animatable,
            DynamicAnimatable
        )

        if isinstance(instance, Animatable):
            return get_bound_method(self, instance)
        if isinstance(instance, DynamicAnimatable):
            return get_dynamic_bound_method(self, instance)
        return self

    @classmethod
    def register(
        cls: type[Self]
    ) -> Callable[
        [Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]],
        Self
    ]:

        def result(
            method: Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]
        ) -> Self:
            assert isinstance(method, classmethod)
            return cls(method.__func__)

        return result

    def iter_animations(
        self: Self,
        dst: AnimatableT,
        *args: P.args,
        **kwargs: P.kwargs
    ) -> Iterator[Animation]:
        return self._method(self._actions_cls, dst, *args, **kwargs)


class DescriptiveAction[ActionsT: Actions, AnimatableT: Animatable, DescriptorParametersT: DescriptorParameters, **P](
    Action[ActionsT, AnimatableT, P]
):
    __slots__ = (
        "_descriptor_parameters_cls",
        "_descriptor_items"
    )

    def __init__(
        self: Self,
        descriptor_parameters_cls: type[DescriptorParametersT],
        method: Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]
    ) -> None:
        super().__init__(method)
        self._descriptor_parameters_cls: type[DescriptorParametersT] = descriptor_parameters_cls
        self._descriptor_items: list[tuple[LazyDescriptor, DescriptorParametersT]] = []

    def iter_descriptor_items(
        self: Self
    ) -> Iterator[tuple[LazyDescriptor, DescriptorParametersT]]:
        yield from self._descriptor_items

    @classmethod
    def register(
        cls: type[Self],
        descriptor_parameters_cls: type[DescriptorParametersT]
    ) -> Callable[
        [Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]],
        Self
    ]:

        def result(
            method: Callable[Concatenate[type[ActionsT], AnimatableT, P], Iterator[Animation]]
        ) -> Self:
            assert isinstance(method, classmethod)
            return cls(descriptor_parameters_cls, method.__func__)

        return result

    def register_descriptor[LazyDescriptorT: LazyDescriptor](
        self: Self,
        **parameters: Any
    ) -> Callable[[LazyDescriptorT], LazyDescriptorT]:

        def result(
            descriptor: LazyDescriptorT
        ) -> LazyDescriptorT:
            assert not descriptor._freeze
            self._descriptor_items.append((descriptor, self._descriptor_parameters_cls(**parameters)))
            return descriptor

        return result



from __future__ import annotations


import itertools
from typing import (
    Iterator,
    Literal,
    Never,
    Self,
    Unpack
)

import attrs
import numpy as np

from ..constants.custom_typing import (
    NP_x2i4,
    NP_x3f8,
    NP_xf8,
    NP_xi4
)
from ..lazy.lazy import Lazy
from .animatable.actions import Action
from .animatable.animatable import (
    Animatable,
    AnimatableActions,
    AnimatableInterpolateAnimation,
    AnimatablePiecewiseAnimation,
    DynamicAnimatable
)
from .animatable.animation import (
    AnimateKwargs,
    Animation
)
from .animatable.piecewiser import Piecewiser


class GraphActions(AnimatableActions):
    __slots__ = ()

    @Action.register()
    @classmethod
    def interpolate(
        cls: type[Self],
        dst: Graph,
        src_0: Graph,
        src_1: Graph
    ) -> Iterator[Animation]:
        yield GraphInterpolateAnimation(dst, src_0, src_1)

    @Action.register()
    @classmethod
    def piecewise(
        cls: type[Self],
        dst: Graph,
        src: Graph,
        piecewiser: Piecewiser
    ) -> Iterator[Animation]:
        yield GraphPiecewiseAnimation(dst, src, piecewiser)


class Graph(GraphActions, Animatable):
    __slots__ = ()

    def __init__(
        self: Self,
        positions: NP_x3f8 | None = None,
        edges: NP_x2i4 | None = None
    ) -> None:
        super().__init__()
        if positions is not None:
            self._positions_ = positions
        if edges is not None:
            self._edges_ = edges

    @Lazy.variable()
    @staticmethod
    def _positions_() -> NP_x3f8:
        return np.zeros((0, 3))

    @Lazy.variable()
    @staticmethod
    def _edges_() -> NP_x2i4:
        return np.zeros((0, 2), dtype=np.int32)

    @Lazy.property()
    @staticmethod
    def _cumlengths_(
        positions: NP_x3f8,
        edges: NP_x2i4
    ) -> NP_xf8:
        lengths = np.linalg.norm(positions[edges[:, 1]] - positions[edges[:, 0]], axis=1)
        return np.insert(np.cumsum(lengths), 0, 0.0)

    def set(
        self: Self,
        positions: NP_x3f8,
        edges: NP_x2i4
    ) -> Self:
        self._positions_ = positions
        self._edges_ = edges
        return self

    def split(
        self: Self,
        dsts: tuple[Self, ...],
        alphas: NP_xf8
    ) -> Self:
        for dst, (positions, edges) in zip(dsts, GraphUtils.graph_split(self, alphas), strict=True):
            dst.set(
                positions=positions,
                edges=edges
            )
        return self

    def concatenate(
        self: Self,
        srcs: tuple[Self, ...]
    ) -> Self:
        positions, edges = GraphUtils.graph_concatenate(srcs)
        self.set(
            positions=positions,
            edges=edges
        )
        return self

    def animate(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> DynamicGraph[Self]:
        return DynamicGraph(self, **kwargs)


class DynamicGraph[GraphT: Graph](GraphActions, DynamicAnimatable[GraphT]):
    __slots__ = ()


@attrs.frozen(kw_only=True)
class GraphInterpolateInfo:
    positions_0: NP_x3f8
    positions_1: NP_x3f8
    edges: NP_x2i4


class GraphInterpolateAnimation[GraphT: Graph](AnimatableInterpolateAnimation[GraphT]):
    __slots__ = ()

    @Lazy.property()
    @staticmethod
    def _interpolate_info_(
        src_0: GraphT,
        src_1: GraphT
    ) -> GraphInterpolateInfo:
        positions_0, positions_1, edges = GraphUtils.graph_interpolate(src_0, src_1)
        return GraphInterpolateInfo(
            positions_0=positions_0,
            positions_1=positions_1,
            edges=edges
        )

    def interpolate(
        self: Self,
        dst: GraphT,
        alpha: float
    ) -> None:
        interpolate_info = self._interpolate_info_
        dst.set(
            positions=(1.0 - alpha) * interpolate_info.positions_0 + alpha * interpolate_info.positions_1,
            edges=interpolate_info.edges
        )


class GraphPiecewiseAnimation[GraphT: Graph](AnimatablePiecewiseAnimation[GraphT]):
    __slots__ = ()

    @classmethod
    def split(
        cls: type[Self],
        dsts: tuple[GraphT, ...],
        src: GraphT,
        alphas: NP_xf8
    ) -> None:
        src.split(dsts, alphas)

    @classmethod
    def concatenate(
        cls: type[Self],
        dst: GraphT,
        srcs: tuple[GraphT, ...]
    ) -> None:
        dst.concatenate(srcs)


class GraphUtils:
    __slots__ = ()

    def __new__(
        cls: type[Self]
    ) -> Never:
        raise TypeError

    @classmethod
    def concatenate_graphs(
        cls: type[Self],
        graphs: Iterator[tuple[NP_x3f8, NP_x2i4]]
    ) -> tuple[NP_x3f8, NP_x2i4]:
        graph_tuple: tuple[tuple[NP_x3f8, NP_x2i4], ...] = (
            (np.zeros((0, 3)), np.zeros((0, 2), dtype=np.int32)),
            *graphs
        )
        offsets = np.roll(np.fromiter((
            len(positions) for positions, _ in graph_tuple
        ), dtype=np.int32).cumsum(), 1)
        return (
            np.concatenate(tuple(positions for positions, _ in graph_tuple)),
            np.concatenate(tuple(edges + offset for (_, edges), offset in zip(graph_tuple, offsets, strict=True)))
        )

    @classmethod
    def graph_split(
        cls: type[Self],
        graph: Graph,
        alphas: NP_xf8
    ) -> Iterator[tuple[NP_x3f8, NP_x2i4]]:
        positions = graph._positions_
        edges = graph._edges_
        cumlengths = graph._cumlengths_
        if not len(edges):
            extended_positions = positions
            piece_edges_tuple = tuple(edges for _ in range(len(alphas) + 1))
        else:
            interpolated_positions, insertion_indices = cls._get_interpolated_samples(
                positions=positions,
                edges=edges,
                knots=cumlengths,
                alphas=alphas * cumlengths[-1]
            )
            extended_positions, extended_edges = cls._insert_samples(
                positions=positions,
                edges=edges,
                interpolated_positions=interpolated_positions,
                insertion_indices=insertion_indices
            )
            piece_edges_tuple = tuple(
                extended_edges[start:stop]
                for start, stop in itertools.pairwise(
                    np.array((0, *(insertion_indices + np.arange(len(alphas)) + 1), len(edges) + len(alphas)))
                )
            )
        for piece_edges in piece_edges_tuple:
            (simplified_positions,), simplified_edges = cls._unify_edges((extended_positions, piece_edges))
            yield simplified_positions, simplified_edges

    @classmethod
    def graph_concatenate(
        cls: type[Self],
        graphs: tuple[Graph, ...]
    ) -> tuple[NP_x3f8, NP_x2i4]:
        (simplified_positions,), simplified_edges = cls._unify_edges(cls.concatenate_graphs(
            (graph._positions_, graph._edges_)
            for graph in graphs
        ))
        return simplified_positions, simplified_edges

    @classmethod
    def graph_interpolate(
        cls: type[Self],
        graph_0: Graph,
        graph_1: Graph
    ) -> tuple[NP_x3f8, NP_x3f8, NP_x2i4]:
        return cls._unify_aligned_graph_pairs(
            *cls._iter_aligned_graph_pairs_for_graph_interpolation(
                positions_0=graph_0._positions_,
                edges_0=graph_0._edges_,
                cumlengths_0=graph_0._cumlengths_,
                positions_1=graph_1._positions_,
                edges_1=graph_1._edges_,
                cumlengths_1=graph_1._cumlengths_
            )
        )

    @classmethod
    def _iter_aligned_graph_pairs_for_graph_interpolation(
        cls: type[Self],
        positions_0: NP_x3f8,
        edges_0: NP_x2i4,
        cumlengths_0: NP_xf8,
        positions_1: NP_x3f8,
        edges_1: NP_x2i4,
        cumlengths_1: NP_xf8
    ) -> Iterator[tuple[tuple[NP_x3f8, NP_x2i4], tuple[NP_x3f8, NP_x2i4]]]:
        match len(edges_0), len(edges_1):
            case 0, 0:
                pass
            case _, 0:
                yield (
                    (positions_0, edges_0),
                    cls._get_centroid_graph(positions_0, edges_0, edges_0)
                )
            case 0, _:
                yield (
                    cls._get_centroid_graph(positions_1, edges_1, edges_1),
                    (positions_1, edges_1)
                )
            case _:
                knots_0 = cumlengths_0 * cumlengths_1[-1]
                knots_1 = cumlengths_1 * cumlengths_0[-1]
                yield (
                    cls._extend_graph_samples(
                        positions=positions_0,
                        edges=edges_0,
                        knots=knots_0,
                        alphas=knots_1[1:-1],
                        side="right"
                    ),
                    cls._extend_graph_samples(
                        positions=positions_1,
                        edges=edges_1,
                        knots=knots_1,
                        alphas=knots_0[1:-1],
                        side="left"
                    )
                )

    @classmethod
    def _unify_aligned_graph_pairs(
        cls: type[Self],
        *aligned_graph_pairs: tuple[tuple[NP_x3f8, NP_x2i4], tuple[NP_x3f8, NP_x2i4]]
    ) -> tuple[NP_x3f8, NP_x3f8, NP_x2i4]:
        aligned_graph_0 = cls.concatenate_graphs(
            graph_0 for graph_0, _ in aligned_graph_pairs
        )
        aligned_graph_1 = cls.concatenate_graphs(
            graph_1 for _, graph_1 in aligned_graph_pairs
        )
        (positions_0, positions_1), edges = cls._unify_edges(aligned_graph_0, aligned_graph_1)
        return positions_0, positions_1, edges

    @classmethod
    def _unify_edges(
        cls: type[Self],
        *aligned_graphs: tuple[NP_x3f8, NP_x2i4]
    ) -> tuple[tuple[NP_x3f8, ...], NP_x2i4]:
        unique_indices_array, inverse = np.unique(
            np.array(tuple(
                edges.flatten()
                for _, edges in aligned_graphs
            )),
            axis=1,
            return_inverse=True
        )
        return (
            tuple(
                positions[unique_indices]
                for (positions, _), unique_indices in zip(aligned_graphs, unique_indices_array, strict=True)
            ),
            inverse.reshape((-1, 2))
        )

    @classmethod
    def _get_centroid_graph(
        cls: type[Self],
        positions: NP_x3f8,
        edges: NP_x2i4,
        aligned_edges: NP_x2i4
    ) -> tuple[NP_x3f8, NP_x2i4]:
        samples = positions[edges.flatten()]
        centroid = (np.max(samples, axis=0) + np.min(samples, axis=0)) / 2.0
        return centroid[None], np.zeros_like(aligned_edges)

    @classmethod
    def _extend_graph_samples(
        cls: type[Self],
        positions: NP_x3f8,
        edges: NP_x2i4,
        knots: NP_xf8,
        alphas: NP_xf8,
        side: Literal["left", "right"]
    ) -> tuple[NP_x3f8, NP_x2i4]:
        interpolated_positions, insertion_indices = cls._get_interpolated_samples(
            positions=positions,
            edges=edges,
            knots=knots,
            alphas=alphas,
            side=side
        )
        return cls._insert_samples(
            positions=positions,
            edges=edges,
            interpolated_positions=interpolated_positions,
            insertion_indices=insertion_indices
        )

    @classmethod
    def _get_interpolated_samples(
        cls: type[Self],
        positions: NP_x3f8,
        edges: NP_x2i4,
        knots: NP_xf8,
        alphas: NP_xf8,
        side: Literal["left", "right"] = "left"
    ) -> tuple[NP_x3f8, NP_xi4]:
        insertion_indices = np.searchsorted(knots[1:-1], alphas, side=side).astype(np.int32)
        residues = (alphas - knots[insertion_indices]) / np.maximum(knots[insertion_indices + 1] - knots[insertion_indices], 1e-8)
        interpolated_positions = (
            (1.0 - residues[:, None]) * positions[edges[insertion_indices, 0]]
            + residues[:, None] * positions[edges[insertion_indices, 1]]
        )
        return interpolated_positions, insertion_indices

    @classmethod
    def _insert_samples(
        cls: type[Self],
        positions: NP_x3f8,
        edges: NP_x2i4,
        interpolated_positions: NP_x3f8,
        insertion_indices: NP_xi4
    ) -> tuple[NP_x3f8, NP_x2i4]:
        interpolated_position_indices = np.arange(len(insertion_indices)) + len(positions)
        return (
            np.concatenate((positions, interpolated_positions)),
            np.column_stack((
                np.insert(edges[:, 0], insertion_indices + 1, interpolated_position_indices),
                np.insert(edges[:, 1], insertion_indices, interpolated_position_indices)
            ))
        )



from __future__ import annotations


from typing import Self

from ...mobjects.mobject import Mobject
from ..composition.parallel import Parallel
from ..timeline import Timeline
from .fade_in import FadeIn
from .fade_out import FadeOut


class FadeTransform(Timeline):
    __slots__ = (
        "_start_mobject",
        "_stop_mobject"
    )

    def __init__(
        self: Self,
        start_mobject: Mobject,
        stop_mobject: Mobject
    ) -> None:
        super().__init__(run_alpha=1.0)
        self._start_mobject: Mobject = start_mobject
        self._stop_mobject: Mobject = stop_mobject

    async def construct(
        self: Self
    ) -> None:
        start_mobject = self._start_mobject
        stop_mobject = self._stop_mobject

        self.scene.add(stop_mobject)
        await self.play(Parallel(
            FadeOut(start_mobject),
            start_mobject.animate().move_to(stop_mobject),
            FadeIn(stop_mobject),
            stop_mobject.animate(rewind=True).move_to(start_mobject)
        ))
        self.scene.discard(start_mobject)



from __future__ import annotations


from ...lazy.lazy import Lazy
from ...toplevel.toplevel import Toplevel
from ..animatable.animatable import AnimatableActions
from ..arrays.animatable_color import AnimatableColor
from ..model import ModelActions
from ..point import Point


class AmbientLight(Point):
    __slots__ = ()

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableColor)
    @Lazy.volatile()
    @staticmethod
    def _color_() -> AnimatableColor:
        return AnimatableColor(Toplevel._get_config().default_color)



from __future__ import annotations


from typing import Self

import moderngl

from ...lazy.lazy import Lazy
from .buffer import Buffer


class TextureBuffer(Buffer):
    __slots__ = ()

    def __init__(
        self: Self,
        *,
        name: str,
        # Note, each texture should occur only once.
        textures: moderngl.Texture | tuple[moderngl.Texture, ...],
        array_lens: dict[str, int] | None = None
    ) -> None:
        if isinstance(textures, tuple):
            shape = (len(textures),)
        else:
            shape = ()
            textures = (textures,)
        super().__init__(
            shape=shape,
            array_lens=array_lens
        )
        self._name_ = name
        self._textures_ = textures

    @Lazy.variable()
    @staticmethod
    def _name_() -> str:
        return ""

    @Lazy.variable(plural=True)
    @staticmethod
    def _textures_() -> tuple[moderngl.Texture, ...]:
        return ()



from __future__ import annotations


import pathlib
import subprocess
from typing import (
    IO,
    Iterator,
    Self
)

import ffmpeg
import numpy as np
from PIL import Image

from ..rendering.buffers.attributes_buffer import AttributesBuffer
from ..rendering.buffers.texture_buffer import TextureBuffer
from ..rendering.framebuffers.final_framebuffer import FinalFramebuffer
from ..rendering.framebuffers.oit_framebuffer import OITFramebuffer
from ..rendering.mgl_enums import PrimitiveMode
from ..rendering.vertex_array import VertexArray
from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


class VideoPipe:
    __slots__ = (
        "_video_path",
        "_video_stream",
        "_writing_process",
        "_writing"
    )

    def __init__(
        self: Self,
        video_path: pathlib.Path
    ) -> None:
        super().__init__()
        writing_process = (
            ffmpeg
            .input(
                filename="pipe:",
                format="rawvideo",
                pix_fmt="rgb24",
                s=f"{Toplevel._get_config().pixel_width}x{Toplevel._get_config().pixel_height}",
                framerate=Toplevel._get_config().fps
            )
            .vflip()
            .output(
                filename=video_path,
                pix_fmt="yuv420p",
                loglevel="quiet"
            )
            .run_async(
                pipe_stdin=True,
                overwrite_output=True
            )
        )
        self._video_path: pathlib.Path = video_path
        self._video_stream: IO[bytes] = writing_process.stdin
        self._writing_process: subprocess.Popen[bytes] = writing_process
        self._writing: bool = False

    @property
    def is_writing(
        self: Self
    ) -> bool:
        return self._writing

    def enable_writing(
        self: Self
    ) -> None:
        if self._writing:
            return
        self._writing = True
        Toplevel._get_logger().log(f"Start recording video to '{self._video_path}'.")

    def disable_writing(
        self: Self
    ) -> None:
        if not self._writing:
            return
        self._writing = False
        Toplevel._get_logger().log(f"Stop recording video to '{self._video_path}'.")

    def write(
        self: Self,
        data: bytes
    ) -> None:
        self._video_stream.write(data)

    def save(
        self: Self
    ) -> None:
        self._writing_process.communicate()
        Toplevel._get_logger().log(f"Recording saved to '{self._video_path}'.")


class Livestreamer:
    __slots__ = ("_livestreaming",)

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        self._livestreaming: bool = False

    @property
    def is_livestreaming(
        self: Self
    ) -> bool:
        return self._livestreaming

    def enable_livestreaming(
        self: Self
    ) -> None:
        if self._livestreaming:
            return
        self._livestreaming = True
        Toplevel._get_logger().log("Start livestreaming.")

    def disable_livestreaming(
        self: Self
    ) -> None:
        if not self._livestreaming:
            return
        self._livestreaming = False
        Toplevel._get_logger().log("Stop livestreaming.")

    def livestream_frame(
        self: Self,
        framebuffer: FinalFramebuffer
    ) -> None:
        Toplevel._get_context().blit_framebuffer(
            dst=Toplevel._get_context().screen_framebuffer,
            src=framebuffer._framebuffer
        )
        Toplevel._get_window()._pyglet_window.flip()


class VideoRecorder:
    __slots__ = (
        "_video_dir",
        "_video_pipes"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        video_dir = Toplevel._get_config().video_output_dir
        video_dir.mkdir(exist_ok=True)
        self._video_dir: pathlib.Path = video_dir
        self._video_pipes: dict[str, VideoPipe] = {}

    @property
    def is_recording(
        self: Self
    ) -> bool:
        return any(video_pipe.is_writing for video_pipe in self._video_pipes.values())

    def enable_recording(
        self: Self,
        filename: str
    ) -> None:
        if (video_pipe := self._video_pipes.get(filename)) is None:
            video_path = self._video_dir.joinpath(filename)
            assert video_path.suffix == ".mp4", \
                f"Video format other than .mp4 is currently not supported: {video_path.suffix}"
            video_pipe = VideoPipe(video_path)
            self._video_pipes[filename] = video_pipe
        video_pipe.enable_writing()

    def disable_recording(
        self: Self,
        filename: str
    ) -> None:
        if (video_pipe := self._video_pipes.get(filename)) is None:
            raise ValueError(f"Video pipe to {filename} not found.")
        video_pipe.disable_writing()

    def record_frame(
        self: Self,
        framebuffer: FinalFramebuffer
    ) -> None:
        frame_data = framebuffer._framebuffer.read()
        for video_pipe in self._video_pipes.values():
            if video_pipe.is_writing:
                video_pipe.write(frame_data)

    def save_videos(
        self: Self
    ) -> None:
        for video_pipe in self._video_pipes.values():
            video_pipe.save()


class ImageRecoder:
    __slots__ = ("_image_dir",)

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        image_dir = Toplevel._get_config().image_output_dir
        image_dir.mkdir(exist_ok=True)
        self._image_dir: pathlib.Path = image_dir

    def record_frame(
        self: Self,
        filename: str,
        framebuffer: FinalFramebuffer
    ) -> None:
        image_path = self._image_dir.joinpath(filename)
        Image.frombytes(
            "RGB",
            Toplevel._get_config().pixel_size,
            framebuffer._framebuffer.read(),
            "raw"
        ).transpose(Image.Transpose.FLIP_TOP_BOTTOM).save(image_path)
        Toplevel._get_logger().log(f"Snapshot saved to '{image_path}'.")


class Renderer(ToplevelResource):
    __slots__ = (
        "_use_msaa",
        "_final_framebuffer",
        "_oit_framebuffer",
        "_oit_msaa_framebuffer",
        "_oit_compose_vertex_array",
        "_livestreamer",
        "_video_recorder",
        "_image_recoder"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()

        msaa_samples = Toplevel._get_config().msaa_samples
        use_msaa = bool(msaa_samples)
        final_framebuffer = FinalFramebuffer()
        oit_framebuffer = OITFramebuffer()
        oit_msaa_framebuffer = OITFramebuffer(samples=msaa_samples) if use_msaa else oit_framebuffer
        oit_compose_vertex_array = VertexArray(
            shader_filename="oit_compose.glsl",
            texture_buffers=(
                TextureBuffer(
                    name="t_accum_map",
                    textures=oit_framebuffer.accum_attachment
                ),
                TextureBuffer(
                    name="t_revealage_map",
                    textures=oit_framebuffer.revealage_attachment
                )
            ),
            attributes_buffer=AttributesBuffer(
                field_declarations=(
                    "vec3 in_position",
                    "vec2 in_uv"
                ),
                data_dict={
                    "in_position": np.array((
                        (-1.0, -1.0, 0.0),
                        (1.0, -1.0, 0.0),
                        (1.0, 1.0, 0.0),
                        (-1.0, 1.0, 0.0)
                    )),
                    "in_uv": np.array((
                        (0.0, 0.0),
                        (1.0, 0.0),
                        (1.0, 1.0),
                        (0.0, 1.0)
                    ))
                },
                primitive_mode=PrimitiveMode.TRIANGLE_FAN,
                vertices_count=4
            )
        )

        self._use_msaa: bool = use_msaa
        self._final_framebuffer: FinalFramebuffer = final_framebuffer
        self._oit_framebuffer: OITFramebuffer = oit_framebuffer
        self._oit_msaa_framebuffer: OITFramebuffer = oit_msaa_framebuffer
        self._oit_compose_vertex_array: VertexArray = oit_compose_vertex_array
        self._livestreamer: Livestreamer = Livestreamer()
        self._video_recorder: VideoRecorder = VideoRecorder()
        self._image_recoder: ImageRecoder = ImageRecoder()

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        Toplevel._renderer = self
        Toplevel._get_config().output_dir.mkdir(exist_ok=True)
        yield
        self._video_recorder.save_videos()
        Toplevel._renderer = None

    def _render_frame(
        self: Self
    ) -> None:
        scene = Toplevel._get_scene()

        self._oit_msaa_framebuffer.clear()
        for mobject in scene._root_mobject.iter_descendants():
            for vertex_array in mobject._iter_vertex_arrays():
                self._oit_msaa_framebuffer.render(vertex_array)
        if self._use_msaa:
            self._oit_framebuffer.copy_from(self._oit_msaa_framebuffer)  # TODO: resolve in final fbo, using sampler2DMS, texelFetch.

        self._final_framebuffer.clear(color=(*scene._background_color, scene._background_opacity))
        self._final_framebuffer.render(self._oit_compose_vertex_array)

    def process_frame(
        self: Self
    ) -> None:
        if Toplevel._get_window()._pyglet_window.context is None:
            # User has attempted to close the window.
            raise KeyboardInterrupt
        if self._livestreamer.is_livestreaming or self._video_recorder.is_recording:
            self._render_frame()
        if self._livestreamer.is_livestreaming:
            self._livestreamer.livestream_frame(self._final_framebuffer)
        if self._video_recorder.is_recording:
            self._video_recorder.record_frame(self._final_framebuffer)

    def start_livestream(
        self: Self
    ) -> None:
        self._livestreamer.enable_livestreaming()

    def stop_livestream(
        self: Self
    ) -> None:
        self._livestreamer.disable_livestreaming()

    def start_recording(
        self: Self,
        filename: str | None
    ) -> None:
        if filename is None:
            filename = f"{Toplevel._get_config().default_filename}.mp4"
        self._video_recorder.enable_recording(filename)

    def stop_recording(
        self: Self,
        filename: str | None
    ) -> None:
        if filename is None:
            filename = f"{Toplevel._get_config().default_filename}.mp4"
        self._video_recorder.disable_recording(filename)

    def snapshot(
        self: Self,
        filename: str | None = None
    ) -> None:
        if filename is None:
            filename = f"{Toplevel._get_config().default_filename}.png"
        self._render_frame()
        self._image_recoder.record_frame(filename, self._final_framebuffer)



from __future__ import annotations


from typing import (
    Iterator,
    Self
)

import pyglet

from .events import (
    Event,
    EventInfo,
    KeyPress,
    KeyRelease,
    MouseDrag,
    MouseMotion,
    MousePress,
    MouseRelease,
    MouseScroll
)
from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


class WindowHandlers:
    __slots__ = ("__weakref__",)

    def on_key_press(
        self: Self,
        symbol: int,
        modifiers: int
    ) -> None:
        Toplevel._get_window().push_event_info(KeyPress(
            symbol=symbol,
            modifiers=modifiers
        )._event_info)

    def on_key_release(
        self: Self,
        symbol: int,
        modifiers: int
    ) -> None:
        Toplevel._get_window().push_event_info(KeyRelease(
            symbol=symbol,
            modifiers=modifiers
        )._event_info)

    def on_mouse_motion(
        self: Self,
        x: int,
        y: int,
        dx: int,
        dy: int
    ) -> None:
        Toplevel._get_window().push_event_info(MouseMotion(
            x=x,
            y=y,
            dx=dx,
            dy=dy
        )._event_info)

    def on_mouse_drag(
        self: Self,
        x: int,
        y: int,
        dx: int,
        dy: int,
        buttons: int,
        modifiers: int
    ) -> None:
        Toplevel._get_window().push_event_info(MouseDrag(
            x=x,
            y=y,
            dx=dx,
            dy=dy,
            buttons=buttons,
            modifiers=modifiers
        )._event_info)

    def on_mouse_press(
        self: Self,
        x: int,
        y: int,
        buttons: int,
        modifiers: int
    ) -> None:
        Toplevel._get_window().push_event_info(MousePress(
            x=x,
            y=y,
            buttons=buttons,
            modifiers=modifiers
        )._event_info)

    def on_mouse_release(
        self: Self,
        x: int,
        y: int,
        buttons: int,
        modifiers: int
    ) -> None:
        Toplevel._get_window().push_event_info(MouseRelease(
            x=x,
            y=y,
            buttons=buttons,
            modifiers=modifiers
        )._event_info)

    def on_mouse_scroll(
        self: Self,
        x: int,
        y: int,
        scroll_x: float,
        scroll_y: float
    ) -> None:
        Toplevel._get_window().push_event_info(MouseScroll(
            x=x,
            y=y,
            scroll_x=scroll_x,
            scroll_y=scroll_y
        )._event_info)

    def on_close(
        self: Self
    ) -> None:
        Toplevel._get_window()._pyglet_window.close()


class Window(ToplevelResource):
    __slots__ = (
        "_pyglet_window",
        "_event_info_queue",
        "_window_handlers"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        window_width, window_height = Toplevel._get_config().window_pixel_size
        major_version, minor_version = Toplevel._get_config().gl_version
        pyglet_window = pyglet.window.Window(
            width=window_width,
            height=window_height,
            config=pyglet.gl.Config(
                double_buffer=True,
                major_version=major_version,
                minor_version=minor_version
            )
        )
        window_handlers = WindowHandlers()
        pyglet_window.push_handlers(window_handlers)

        self._pyglet_window: pyglet.window.Window = pyglet_window
        self._event_info_queue: list[EventInfo] = []
        # Keep a strong reference to the handler object, according to
        # `https://pyglet.readthedocs.io/en/latest/programming_guide/events.html#stacking-event-handlers`.
        self._window_handlers: WindowHandlers = window_handlers

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        Toplevel._window = self
        yield
        self._pyglet_window.close()
        Toplevel._window = None

    def push_event_info(
        self: Self,
        event_info: EventInfo
    ) -> None:
        self._event_info_queue.append(event_info)

    def capture_event_info_by(
        self: Self,
        event: Event
    ) -> bool:
        event_info_queue = self._event_info_queue
        for event_info in event_info_queue:
            if event._capture(event_info):
                event_info_queue.remove(event_info)
                return True
        return False

    def clear_event_info_queue(
        self: Self
    ) -> None:
        self._event_info_queue.clear()



from __future__ import annotations


from typing import Self

from .regular_polygon import RegularPolygon


class Circle(RegularPolygon):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        super().__init__(n=64)



from __future__ import annotations


from typing import Self

import numpy as np
from scipy.spatial.transform import Rotation


from ..constants.constants import (
    DL,
    ORIGIN,
    OUT,
    RIGHT,
    UP,
    UR
)
from ..constants.custom_typing import (
    NP_2f8,
    NP_3f8,
    NP_44f8,
    NP_f8,
    NP_x3f8
)
from ..lazy.lazy import Lazy
from ..rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from ..toplevel.toplevel import Toplevel
from .animatable.animatable import AnimatableActions
from .arrays.animatable_float import AnimatableFloat
from .arrays.model_matrix import ModelMatrix
from .model import Model


class Camera(Model):
    __slots__ = ()

    def __init__(
        self: Self,
        near: float | None = None,
        far: float | None = None
    ) -> None:
        super().__init__()
        # Positions bound to `model_matrix`:
        # `target`: ORIGIN
        # `eye`: OUT
        # Distances bound to `model_matrix`:
        # `frame_radii`: (|RIGHT - ORIGIN|, |UP - ORIGIN|)
        # `distance`: |OUT - ORIGIN|
        self.scale(np.append(
            Toplevel._get_config().frame_radii,
            Toplevel._get_config().camera_distance
        ))
        if near is not None:
            self._near_ = AnimatableFloat(near)
        if far is not None:
            self._far_ = AnimatableFloat(far)

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _near_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().camera_near)

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _far_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().camera_far)

    @Lazy.property()
    @staticmethod
    def _target_(
        model_matrix__array: NP_44f8
    ) -> NP_3f8:
        return ModelMatrix._apply(model_matrix__array, ORIGIN)

    @Lazy.property()
    @staticmethod
    def _eye_(
        model_matrix__array: NP_44f8
    ) -> NP_3f8:
        return ModelMatrix._apply(model_matrix__array, OUT)

    @Lazy.property()
    @staticmethod
    def _frame_radii_(
        model_matrix__array: NP_44f8,
        target: NP_3f8
    ) -> NP_2f8:
        return np.array((
            np.linalg.norm(ModelMatrix._apply(model_matrix__array, RIGHT) - target),
            np.linalg.norm(ModelMatrix._apply(model_matrix__array, UP) - target)
        ))

    @Lazy.property()
    @staticmethod
    def _distance_(
        eye: NP_3f8,
        target: NP_3f8
    ) -> NP_f8:
        return np.array(np.linalg.norm(eye - target))

    @Lazy.property()
    @staticmethod
    def _projection_matrix_(
        frame_radii: NP_2f8,
        near__array: NP_f8,
        far__array: NP_f8,
        distance: NP_f8
    ) -> NP_44f8:
        near = near__array
        far = far__array
        sx, sy = distance / frame_radii
        sz = -(far + near) / (far - near)
        tz = -2.0 * far * near / (far - near)
        return np.array((
            ( sx, 0.0,  0.0, 0.0),
            (0.0,  sy,  0.0, 0.0),
            (0.0, 0.0,   sz,  tz),
            (0.0, 0.0, -1.0, 0.0)
        ))

    @Lazy.property()
    @staticmethod
    def _view_matrix_(
        model_matrix__array: NP_44f8,
        eye: NP_3f8
    ) -> NP_44f8:
        model_basis = model_matrix__array[:3, :3]
        model_basis_normalized = model_basis / np.linalg.norm(model_basis, axis=0, keepdims=True)
        return (
            ModelMatrix._matrix_from_rotate(-Rotation.from_matrix(model_basis_normalized).as_rotvec())
            @ ModelMatrix._matrix_from_shift(-eye)
        )

    @Lazy.property()
    @staticmethod
    def _camera_uniform_block_buffer_(
        projection_matrix: NP_44f8,
        view_matrix: NP_44f8,
        frame_radii: NP_2f8
    ) -> UniformBlockBuffer:
        return UniformBlockBuffer(
            name="ub_camera",
            field_declarations=(
                "mat4 u_projection_matrix",
                "mat4 u_view_matrix",
                "vec2 u_frame_radii"
            ),
            data_dict={
                "u_projection_matrix": projection_matrix.T,
                "u_view_matrix": view_matrix.T,
                "u_frame_radii": frame_radii
            }
        )

    @Lazy.property()
    @staticmethod
    def _local_sample_positions_() -> NP_x3f8:
        # Form a rectangle covering the screen,
        # so that aligning on border can be achieved by aligning with the camera.
        return np.array((UR, DL))



from __future__ import annotations


from typing import Self

import numpy as np

from .parametric_surface import ParametricSurface


class Plane(ParametricSurface):
    __slots__ = ()

    def __init__(
        self: Self
    ) -> None:
        super().__init__(
            func=lambda samples: np.concatenate((samples, np.zeros((len(samples), 1))), axis=1),
            normal_func=lambda samples: np.concatenate((np.zeros_like(samples), np.ones((len(samples), 1))), axis=1),
            u_range=(-1.0, 1.0),
            v_range=(-1.0, 1.0),
            resolution=(1, 1)
        )



from __future__ import annotations


from typing import Self

from ...mobjects.mobject import Mobject
from ..timeline import Timeline


class FadeOut(Timeline):
    __slots__ = ("_mobject",)

    def __init__(
        self: Self,
        mobject: Mobject
    ) -> None:
        super().__init__(run_alpha=1.0)
        self._mobject: Mobject = mobject

    async def construct(
        self: Self
    ) -> None:
        mobject = self._mobject

        await self.play(mobject.animate().set(opacity=0.0))
        self.scene.discard(mobject)



from __future__ import annotations


import itertools
from typing import (
    Iterator,
    Self
)


from ...constants.custom_typing import SelectorType
from ...mobjects.string_mobjects.string_mobject import StringMobject
from ..composition.parallel import Parallel
from ..timeline import Timeline
from ..fade.fade_in import FadeIn
from ..fade.fade_out import FadeOut
from .transform import Transform


class TransformMatchingStrings(Timeline):
    __slots__ = (
        "_start_mobject",
        "_stop_mobject",
        "_matched_indices_tuples"
    )

    def __init__(
        self: Self,
        start_mobject: StringMobject,
        stop_mobject: StringMobject,
        key_map: dict[SelectorType, SelectorType] | None = None
    ) -> None:

        def match_elements_evenly[T](
            elements_0: tuple[T, ...],
            elements_1: tuple[T, ...]
        ) -> Iterator[tuple[tuple[T, ...], tuple[T, ...]]]:
            len_0 = len(elements_0)
            len_1 = len(elements_1)
            if len_0 > len_1:
                q, r = divmod(len_0, len_1)
                iter_0 = itertools.chain(range(0, r * (q + 1), q + 1), range(r * (q + 1), len_0 + 1, q))
                iter_1 = iter(range(len_1 + 1))
            else:
                q, r = divmod(len_1, len_0)
                iter_0 = iter(range(len_0 + 1))
                iter_1 = itertools.chain(range(0, r * (q + 1), q + 1), range(r * (q + 1), len_1 + 1, q))
            for (start_0, start_1), (stop_0, stop_1) in itertools.pairwise(zip(iter_0, iter_1, strict=True)):
                yield elements_0[start_0:stop_0], elements_1[start_1:stop_1]

        if key_map is None:
            key_map = {}

        matched_indices_tuples: list[tuple[tuple[tuple[int, ...], ...], tuple[tuple[int, ...], ...]]] = []
        indices_0: set[int] = set()
        indices_1: set[int] = set()
        for selector_0, selector_1 in itertools.chain(
            key_map.items(),
            (
                (matched_key, matched_key)
                for matched_key in sorted(
                    set.intersection(
                        set(start_mobject._string[span.as_slice()] for span in start_mobject._isolated_spans),
                        set(stop_mobject._string[span.as_slice()] for span in stop_mobject._isolated_spans)
                    ),
                    key=len
                )
            )
        ):
            indices_tuple_0 = tuple(
                indices
                for indices in start_mobject._get_indices_tuple_by_selector(selector_0)
                if indices_0.isdisjoint(indices)
            )
            indices_tuple_1 = tuple(
                indices
                for indices in stop_mobject._get_indices_tuple_by_selector(selector_1)
                if indices_1.isdisjoint(indices)
            )
            if not indices_tuple_0 or not indices_tuple_1:
                continue
            indices_0.update(itertools.chain.from_iterable(indices_tuple_0))
            indices_1.update(itertools.chain.from_iterable(indices_tuple_1))
            matched_indices_tuples.extend(match_elements_evenly(indices_tuple_0, indices_tuple_1))

        super().__init__(run_alpha=1.0)
        self._start_mobject: StringMobject = start_mobject
        self._stop_mobject: StringMobject = stop_mobject
        self._matched_indices_tuples: tuple[tuple[tuple[tuple[int, ...], ...], tuple[tuple[int, ...], ...]], ...] = tuple(
            matched_indices_tuples
        )

    async def construct(
        self: Self
    ) -> None:
        start_mobject = self._start_mobject
        stop_mobject = self._stop_mobject
        matched_indices_tuples = self._matched_indices_tuples
        matched_mobjects_pairs = tuple(
            (start_mobject._shape_mobjects[index_0].copy(), stop_mobject._shape_mobjects[index_1].copy())
            for indices_tuples_0, indices_tuples_1 in matched_indices_tuples
            for indices_0, indices_1 in itertools.product(indices_tuples_0, indices_tuples_1)
            for index_0, index_1 in zip(indices_0, indices_1, strict=True)
        )
        mismatched_mobjects_0 = tuple(
            start_mobject._shape_mobjects[index_0].copy()
            for index_0 in set(range(len(start_mobject._shape_mobjects))).difference(*itertools.chain.from_iterable(
                indices_tuples_0
                for indices_tuples_0, _ in matched_indices_tuples
            ))
        )
        mismatched_mobjects_1 = tuple(
            stop_mobject._shape_mobjects[index_1].copy()
            for index_1 in set(range(len(stop_mobject._shape_mobjects))).difference(*itertools.chain.from_iterable(
                indices_tuples_1
                for _, indices_tuples_1 in matched_indices_tuples
            ))
        )

        self.scene.discard(start_mobject)
        self.scene.add(
            *(matched_mobject_0 for matched_mobject_0, _ in matched_mobjects_pairs),
            *mismatched_mobjects_0
        )
        await self.play(Parallel(
            *(
                Transform(matched_mobject_0, matched_mobject_1)
                for matched_mobject_0, matched_mobject_1 in matched_mobjects_pairs
            ),
            *(
                FadeOut(mismatched_mobject_0)
                for mismatched_mobject_0 in mismatched_mobjects_0
            ),
            *(
                FadeIn(mismatched_mobject_1)
                for mismatched_mobject_1 in mismatched_mobjects_1
            )
        ))
        self.scene.discard(
            *(matched_mobject_1 for _, matched_mobject_1 in matched_mobjects_pairs),
            *mismatched_mobjects_1
        )
        self.scene.add(stop_mobject)



from __future__ import annotations


import hashlib
import json
import pathlib
from abc import (
    ABC,
    abstractmethod
)
from typing import (
    Never,
    Self,
    TypedDict
)

import attrs

from ..toplevel.toplevel import Toplevel


@attrs.frozen(kw_only=True)
class MobjectInput:
    pass


@attrs.frozen(kw_only=True)
class MobjectOutput:
    pass


class MobjectJSON(TypedDict):
    pass


class MobjectIO[MobjectInputT: MobjectInput, MobjectOutputT: MobjectOutput, MobjectJSONT: MobjectJSON](ABC):
    __slots__ = ()

    def __new__(
        cls: type[Self]
    ) -> Never:
        raise TypeError

    @classmethod
    def get(
        cls: type[Self],
        input_data: MobjectInputT
    ) -> MobjectOutputT:
        # Notice that as we are using string as key,
        # each item shall have an explicit string representation of data,
        # which shall not contain any information varying in each run, like addresses.
        hash_content = f"{input_data}"
        # Truncating at 16 bytes for cleanliness.
        hex_string = hashlib.sha256(hash_content.encode()).hexdigest()[:16]
        json_path = cls._get_output_subdir(cls._get_subdir_name()).joinpath(f"{hex_string}.json")
        if json_path.exists():
            Toplevel._get_logger().log(f"Using cached intermediate files in {cls.__name__}.")
        else:
            Toplevel._get_logger().log(f"Generating intermediate files in {cls.__name__}...")
            temp_path = cls._get_output_subdir("_temp").joinpath(hex_string)
            output_data = cls.generate(input_data, temp_path)
            json_data = cls.dump_json(output_data)
            json_text = json.dumps(json_data, ensure_ascii=False)
            json_path.write_text(json_text, encoding="utf-8")
            Toplevel._get_logger().log(f"Intermediate files generation completed.")
        json_text = json_path.read_text(encoding="utf-8")
        json_data = json.loads(json_text)
        return cls.load_json(json_data)

    @classmethod
    def _get_output_subdir(
        cls: type[Self],
        dir_name: str
    ) -> pathlib.Path:
        subdir = Toplevel._get_config().output_dir.joinpath(dir_name)
        subdir.mkdir(exist_ok=True)
        return subdir

    @classmethod
    @abstractmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        pass

    @classmethod
    @abstractmethod
    def generate(
        cls: type[Self],
        input_data: MobjectInputT,
        temp_path: pathlib.Path
    ) -> MobjectOutputT:
        pass

    @classmethod
    @abstractmethod
    def dump_json(
        cls: type[Self],
        output_data: MobjectOutputT
    ) -> MobjectJSONT:
        pass

    @classmethod
    @abstractmethod
    def load_json(
        cls: type[Self],
        json_data: MobjectJSONT
    ) -> MobjectOutputT:
        pass



from __future__ import annotations


__all__ = (
    "NP_f8",
    "NP_2f8",
    "NP_3f8",
    "NP_44f8",
    "NP_xf8",
    "NP_x2f8",
    "NP_x3f8",
    "NP_xi4",
    "NP_x2i4",
    "NP_x3i4",
    "NP_xxi4",
    "ShapeType",
    "ColorType",
    "AlignmentType",
    "SelectorType",
    "RateType",
    "ConditionType"
)


import re
from typing import (
    Callable,
    Literal
)

import numpy as np
from colour import Color


type _XD = int
type _2D = Literal[2]
type _3D = Literal[3]
type _4D = Literal[4]

type NP_f8 = np.ndarray[tuple[()], np.dtype[np.float64]]
type NP_2f8 = np.ndarray[tuple[_2D], np.dtype[np.float64]]
type NP_3f8 = np.ndarray[tuple[_3D], np.dtype[np.float64]]
type NP_44f8 = np.ndarray[tuple[_4D, _4D], np.dtype[np.float64]]
type NP_xf8 = np.ndarray[tuple[_XD], np.dtype[np.float64]]
type NP_x2f8 = np.ndarray[tuple[_XD, _2D], np.dtype[np.float64]]
type NP_x3f8 = np.ndarray[tuple[_XD, _3D], np.dtype[np.float64]]

type NP_xi4 = np.ndarray[tuple[_XD], np.dtype[np.int32]]
type NP_x2i4 = np.ndarray[tuple[_XD, _2D], np.dtype[np.int32]]
type NP_x3i4 = np.ndarray[tuple[_XD, _3D], np.dtype[np.int32]]
type NP_xxi4 = np.ndarray[tuple[_XD, _XD], np.dtype[np.int32]]

type ShapeType = tuple[int, ...]
type ColorType = Color | str | NP_3f8
type AlignmentType = Literal["left", "center", "right"]
type SelectorType = str | re.Pattern[str] | slice

type RateType = Callable[[float], float]
type ConditionType = Callable[[], bool]



from __future__ import annotations


import pathlib
import sys
from typing import (
    Iterator,
    Self
)

import attrs
from colour import Color

from ..constants.custom_typing import (
    AlignmentType,
    ColorType
)
from .toplevel import Toplevel
from .toplevel_resource import ToplevelResource


@attrs.frozen(kw_only=True)
class Config(ToplevelResource):
    gl_version: tuple[int, int] = (4, 3)
    fps: int = 30
    aspect_ratio: float = 16.0 / 9.0
    frame_height: float = 8.0
    pixel_height: int = 1080
    window_pixel_height: int = 540
    msaa_samples: int = 4  # Set to 0 to turn off msaa.

    default_color: ColorType = Color("white")
    default_opacity: float = 1.0
    default_weight: float = 1.0
    background_color: ColorType = Color("black")
    background_opacity: float = 0.0

    camera_distance: float = 5.0
    camera_near: float = 0.1
    camera_far: float = 100.0

    mesh_ambient_strength: float = 1.0
    mesh_specular_strength: float = 0.5
    mesh_shininess: float = 32.0
    graph_width: float = 0.05

    tex_alignment: AlignmentType = "left"
    tex_template: str = "default"
    tex_compiler: str = "xelatex"
    tex_preambles: tuple[str, ...] = (
        "\\usepackage[UTF8]{ctex}",
        "\\usepackage{amsmath}",
        "\\usepackage{amssymb}",
        "\\usepackage{xcolor}"  # Required for labelling.
    )
    math_tex_alignment: AlignmentType = "center"
    math_tex_inline: bool = False
    # See `https://docs.mathjax.org/en/latest/input/tex/extensions/index.html`.
    mathjax_extensions: tuple[str, ...] = (
        "ams",
        "autoload",
        "base",
        "color",  # Required for labelling.
        "newcommand",
        "require"
    )
    mathjax_alignment: AlignmentType = "center"
    mathjax_inline: bool = False

    pango_alignment: AlignmentType = "left"
    pango_font: str = "Consolas"
    code_font: str = "Consolas"
    code_language_suffix: str = ".py"

    shader_search_dirs: tuple[pathlib.Path, ...] = (
        pathlib.Path(),
        pathlib.Path(__import__("manim3").__file__).parent.joinpath("shaders")
    )
    image_search_dirs: tuple[pathlib.Path, ...] = (
        pathlib.Path(),
    )
    output_dir: pathlib.Path = pathlib.Path("manim3_output")
    video_output_dir: pathlib.Path = pathlib.Path("manim3_output/videos")
    image_output_dir: pathlib.Path = pathlib.Path("manim3_output/images")
    default_filename: str = sys.argv[0].removesuffix(".py")

    @property
    def gl_version_code(
        self: Self
    ) -> int:
        major_version, minor_version = self.gl_version
        return major_version * 100 + minor_version * 10

    @property
    def frame_width(
        self: Self
    ) -> float:
        return self.aspect_ratio * self.frame_height

    @property
    def frame_size(
        self: Self
    ) -> tuple[float, float]:
        return self.frame_width, self.frame_height

    @property
    def frame_radii(
        self: Self
    ) -> tuple[float, float]:
        return self.frame_width / 2.0, self.frame_height / 2.0

    @property
    def pixel_width(
        self: Self
    ) -> int:
        return int(self.aspect_ratio * self.pixel_height)

    @property
    def pixel_size(
        self: Self
    ) -> tuple[int, int]:
        return self.pixel_width, self.pixel_height

    @property
    def pixel_per_unit(
        self: Self
    ) -> int:
        return int(self.pixel_height / self.frame_height)

    @property
    def window_pixel_width(
        self: Self
    ) -> int:
        return int(self.aspect_ratio * self.window_pixel_height)

    @property
    def window_pixel_size(
        self: Self
    ) -> tuple[int, int]:
        return self.window_pixel_width, self.window_pixel_height

    def __contextmanager__(
        self: Self
    ) -> Iterator[None]:
        from .timer import Timer
        from .logger import Logger
        from .window import Window
        from .context import Context
        from .renderer import Renderer

        Toplevel._config = self
        with (
            Timer(),
            Logger(),
            Window(),
            Context(),
            Renderer()
        ):
            yield
        Toplevel._config = None



from __future__ import annotations


from typing import Self

import numpy as np
from scipy.spatial.transform import Rotation

from ...constants.custom_typing import (
    NP_3f8,
    NP_44f8,
    NP_x3f8
)
from ...lazy.lazy import Lazy
from .animatable_array import AnimatableArray


class ModelMatrix(AnimatableArray[NP_44f8]):
    __slots__ = ()

    @Lazy.variable()
    @staticmethod
    def _array_() -> NP_44f8:
        return np.identity(4)

    @classmethod
    def _apply(
        cls: type[Self],
        matrix: NP_44f8,
        vector: NP_3f8
    ) -> NP_3f8:
        v = matrix @ np.append(vector, 1.0)
        w_component = v[-1]
        result = np.delete(v, -1)
        if not np.allclose(w_component, 1.0):
            result /= w_component
        return result

    @classmethod
    def _apply_multiple(
        cls: type[Self],
        matrix: NP_44f8,
        vectors: NP_x3f8
    ) -> NP_x3f8:
        v = matrix @ np.append(vectors.T, np.ones((1, len(vectors))), axis=0)
        w_component = v[-1]
        result = np.delete(v, -1, axis=0)
        if not np.allclose(w_component, 1.0):
            result /= w_component
        return result.T

    @classmethod
    def _matrix_from_shift(
        cls: type[Self],
        vector: NP_3f8
    ) -> NP_44f8:
        matrix = np.identity(4)
        matrix[:3, 3] = vector
        return matrix

    @classmethod
    def _matrix_from_scale(
        cls: type[Self],
        factor: NP_3f8
    ) -> NP_44f8:
        matrix = np.identity(4)
        matrix[:3, :3] = np.diag(factor)
        return matrix

    @classmethod
    def _matrix_from_rotate(
        cls: type[Self],
        rotvec: NP_3f8
    ) -> NP_44f8:
        matrix = np.identity(4)
        matrix[:3, :3] = Rotation.from_rotvec(rotvec).as_matrix()
        return matrix



from __future__ import annotations


import pathlib
import re
import subprocess
from typing import (
    Self,
    Unpack
)

import attrs

from ...constants.custom_typing import AlignmentType
from ...toplevel.toplevel import Toplevel
from .latex_string_mobject import (
    LatexStringMobjectIO,
    LatexStringMobjectInput,
    LatexStringMobjectKwargs
)
from .string_mobject import StringMobject


@attrs.frozen(kw_only=True)
class MathJaxInput(LatexStringMobjectInput):
    alignment: AlignmentType = attrs.field(factory=lambda: Toplevel._get_config().mathjax_alignment)
    extensions: tuple[str, ...] = attrs.field(factory=lambda: Toplevel._get_config().mathjax_extensions)
    inline: bool = attrs.field(factory=lambda: Toplevel._get_config().mathjax_inline)


class MathJaxKwargs(LatexStringMobjectKwargs, total=False):
    alignment: AlignmentType
    extensions: tuple[str, ...]
    inline: bool


class MathJaxIO[MathJaxInputT: MathJaxInput](LatexStringMobjectIO[MathJaxInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "mathjax"

    @classmethod
    def _create_svg(
        cls: type[Self],
        content: str,
        input_data: MathJaxInputT,
        svg_path: pathlib.Path
    ) -> None:
        mathjax_program_path = pathlib.Path(__import__("manim3").__file__).parent.joinpath("plugins/mathjax/index.js")

        if subprocess.run((
            "node",
            mathjax_program_path,
            "--tex", content,
            "--extensions", " ".join(input_data.extensions),
            "--alignment", input_data.alignment,
            "--inline", "true" if input_data.inline else "false",
            "--path", svg_path
        ), stdout=subprocess.DEVNULL).returncode:
            raise OSError("MathJaxIO: Failed to execute node command")
        svg_text = svg_path.read_text(encoding="utf-8")
        if (error_match_obj := re.search(r"<text\b.*?>(.*)</text>", svg_text)) is not None:
            error = OSError("MathJaxIO: Failed to execute node command")
            error.add_note(f"MathJax error: {error_match_obj.group(1)}")
            raise error

        # Seems svgelements cannot handle units in the root tag, so remove them manually.
        svg_path.write_text(re.sub(
            r"(width|height)=\"([\d\.]+)ex\"",
            lambda match: f"{match.group(1)}=\"{match.group(2)}\"",
            svg_path.read_text(encoding="utf-8")
        ), encoding="utf-8")

    @classmethod
    def _get_adjustment_scale(
        cls: type[Self]
    ) -> float:
        return 0.4021


class MathJax(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[MathJaxKwargs]
    ) -> None:
        super().__init__(MathJaxIO.get(MathJaxInput(string=string, **kwargs)))



from __future__ import annotations


from typing import Self

from ..mgl_enums import (
    BlendEquation,
    BlendFunc,
    ContextFlag
)
from .framebuffer import (
    Framebuffer,
    AttachmentInfo
)


class FinalFramebuffer(Framebuffer):
    __slots__ = ()

    def __init__(
        self: Self,
        samples: int = 0
    ) -> None:
        super().__init__(
            attachment_info_dict={
                "": AttachmentInfo(
                    components=4,
                    dtype="f1",
                    src_blend_func=BlendFunc.SRC_ALPHA,
                    dst_blend_func=BlendFunc.ONE_MINUS_SRC_ALPHA,
                    blend_equation=BlendEquation.FUNC_ADD
                )
            },
            samples=samples,
            flag=ContextFlag.BLEND
        )
